import bpy

from .color_ramps import duplicate_color_ramp
from .palettes import copy_palette


class ColorBoxTypeMixin:
    """
    ColorPalettes and ColorRamps have the exact same Collection behaviour for adding and removing collections.
    The type (Palette or Ramp) is determined by the 'collection_type': 0 for palette, 1 for color ramp

    """

    collection_type: bpy.props.IntProperty(default=-1)
    collection_index: bpy.props.IntProperty()
    row_index: bpy.props.IntProperty(default=-1)

    def get_collection_type(self, scene):
        if self.collection_type == 0:
            return scene.color_box
        elif self.collection_type == 1:
            return scene.color_box_ramps

    def get_collection(self, scene):
        return self.get_collection_type(scene)[self.collection_index]


class COLORBOX_OT_collection_add(ColorBoxTypeMixin, bpy.types.Operator):
    bl_idname = "color_box.collection_add"
    bl_label = "Add collection"
    bl_description = "Add a new collection"
    bl_options = {"REGISTER"}

    def execute(self, context):
        scene = context.scene
        new_collection = self.get_collection_type(scene).add()
        new_collection.name = "New Collection"
        new_collection.show = True

        from .color_swapper import sync_color_slots

        sync_color_slots(scene)

        return {"FINISHED"}


class COLORBOX_OT_collection_remove(ColorBoxTypeMixin, bpy.types.Operator):
    bl_idname = "color_box.collection_remove"
    bl_label = "Delete collection"
    bl_options = {"REGISTER"}

    def execute(self, context):
        scene = context.scene

        self.get_collection_type(scene).remove(self.collection_index)
        if self.collection_type == 0:
            from .color_swapper import sync_color_slots

            sync_color_slots(scene)
        self.report({"INFO"}, "Removed Collection")
        return {"FINISHED"}


class COLORBOX_OT_collection_show(ColorBoxTypeMixin, bpy.types.Operator):
    bl_idname = "color_box.collection_show"
    bl_label = "Show / hide collection"
    bl_options = {"REGISTER"}

    def execute(self, context):
        collection = self.get_collection(context.scene)
        collection.show = not collection.show
        return {"FINISHED"}


class COLORBOX_OT_collection_reverse(bpy.types.Operator):
    bl_idname = "color_box.collection_reverse"
    bl_label = "Reverse the order"
    bl_description = "Reverse the order of items in this collection"
    bl_options = {"REGISTER"}

    def execute(self, context):
        scene = context.scene
        _type, col_index, row_index = scene.color_box_selected_menu

        if _type == 0 and col_index >= 0:
            collection = scene.color_box[col_index]

            last_index = len(collection.palettes) - 1
            for i in range(last_index):
                collection.palettes.move(last_index, i)

            context.area.tag_redraw()
        return {"FINISHED"}


class COLORBOX_OT_collection_duplicate(bpy.types.Operator):
    bl_idname = "color_box.collection_duplicate"
    bl_label = "Duplicate collection"
    bl_options = {"REGISTER"}

    def execute(self, context):
        scene = context.scene
        _type, col_index, row_index = scene.color_box_selected_menu

        if _type == 0:
            color_box = scene.color_box
            collection = color_box[col_index]
            new_collection = color_box.add()
            new_collection.name = f"{collection.name}_copy"

            for palette in collection.palettes:
                copy_palette(new_collection, palette)

            color_box.move(len(scene.color_box) - 1, col_index + 1)
        elif _type == 1:
            color_ramps = scene.color_box_ramps

            collection_from = color_ramps[col_index]
            new_collection = color_ramps.add()
            new_collection.name = f"{collection_from.name}_copy"

            for _row_index in range(len(collection_from.ramps)):
                duplicate_color_ramp(
                    collection_from,
                    collection_to=new_collection,
                    row_index=_row_index,
                )

            color_ramps.move(len(color_ramps) - 1, col_index + 1)

        context.area.tag_redraw()
        return {"FINISHED"}
