import math
import sys
import bpy
from pathlib import Path

from ..utils.move_rows import add_palette
from ..utils.image import get_hex_code
from ..utils.shaders import (
    draw_text,
    draw_filled_rectangle,
    draw_2d_line,
    draw_rounded_border,
    draw_filled_rounded_rectangle,
    draw_image,
    draw_rounded_border_perforated,
)

from .. import ADDON_ID
WHITE = (1.0, 1.0, 1.0, 1.0)
BLACK = (0.0, 0.0, 0.0, 1.0)
GRAY_PANEL_COLOR = (0.6, 0.6, 0.6, 1.0)
ADDON_NAME = ADDON_ID

class COLORBOX_OT_modal_color_picker(bpy.types.Operator):
    """ A color picker that only works in the ImageEditor """

    bl_idname = "color_box.modal_color_picker"
    bl_label = "Color Picker"

    collection_index: bpy.props.IntProperty(default=-1)
    row_index: bpy.props.IntProperty(default=-1)

    picked_colors = []
    cursor_on_image = False
    current_color = None
    image_editor_dimensions = None, None

    #
    # Below settings will be set in the invoke
    #
    scale = 1.5
    colors_per_row = 4
    display_live_hexcodes = True
    add_hexcodes_to_colorbox = True
    box_width = 60.0
    border_width = 15

    @classmethod
    def poll(cls, context):
        return context.area.type == "IMAGE_EDITOR"

    def draw_logo_and_help_text(self):
        text_start_x, text_start_y = 10, 10
        draw_text(
            text_start_x,
            text_start_y,
            font_size=int(25 * self.ui_scale),
            text="Pick colors by clicking on the image. Press ESC to cancel. ENTER to save and exit.",
            text_color=(1.0, 0.0, 0.0, 0.0),
        )

        if self.logo_image:
            image_size = int(self.ui_scale * 120.0)
            draw_image(
                self.logo_image,
                location=(text_start_x, text_start_y + int(self.ui_scale * 5)),
                width=image_size,
                height=image_size,
            )

    def draw_cursor(self):
        """
        Draw a custom cursor: a color picker widget, which is a sniper (+ sign),
        surrounded by the active color, and optionally the current hex code displayed underneath it
        """
        # draw color sniper + sign
        outer_border_width = self.widget_outer_border_width
        draw_2d_line(
            coords=[
                (self.mouse_reg_x - self.widget_diff, self.mouse_reg_y),
                (self.mouse_reg_x + self.widget_diff, self.mouse_reg_y),
            ],
            color=BLACK,
        )
        draw_2d_line(
            coords=[
                (self.mouse_reg_x, self.mouse_reg_y - self.widget_diff),
                (self.mouse_reg_x, self.mouse_reg_y + self.widget_diff),
            ],
            color=BLACK,
        )

        start_x = (
            self.mouse_reg_x
            - self.widget_half_box_size
            - self.widget_half_outer_width_diff
        )
        start_y = (
            self.mouse_reg_y
            - self.widget_half_box_size
            - self.widget_half_outer_width_diff
            - self.widget_hex_text_height
            + 3
        )  # ??
        width = self.box_width + self.widget_outer_width_diff

        black_border_width = self.scale
        double_black_border = black_border_width * 2
        draw_rounded_border(
            start_x=start_x - black_border_width,
            start_y=start_y - black_border_width,
            width=width + double_black_border,
            height=self.box_width
            + outer_border_width
            + self.widget_hex_text_height
            + double_black_border,
            color=BLACK,
            rounding_radius=outer_border_width + black_border_width,
            segments=24,
        )
        draw_filled_rectangle(
            start_x=start_x - black_border_width,
            start_y=start_y + outer_border_width - black_border_width,
            width=width + double_black_border,
            height=self.widget_hex_text_height + double_black_border,
            color=BLACK,
        )

        draw_rounded_border(
            start_x=start_x,
            start_y=start_y,
            width=width,
            height=self.box_width
            + outer_border_width
            + self.widget_hex_text_height,
            color=GRAY_PANEL_COLOR,
            rounding_radius=outer_border_width,
            segments=24,
        )

        # if self.display_live_hexcodes:
        draw_filled_rectangle(
            start_x=start_x,
            start_y=start_y + outer_border_width,
            width=width,
            height=self.widget_hex_text_height,
            # color=(1.0, 0.0, 0.0, 1.0)  #grey_panel_color
            color=GRAY_PANEL_COLOR,
        )

        if self.current_color is not None:
            draw_rounded_border_perforated(
                start_x=self.mouse_reg_x - self.widget_half_box_size,
                start_y=self.mouse_reg_y - self.widget_half_box_size,
                width=self.box_width,
                height=self.box_width,
                color=self.current_color,
                rounding_radius=self.widget_inner_border_width * 1.2,
                segments=24,
            )

            if self.display_live_hexcodes:
                draw_text(
                    start_x
                    + self.widget_offset
                    + self.widget_half_outer_border_width,
                    start_y
                    + self.widget_half_outer_border_width,  # the text can take a bit of the outer border width
                    font_size=self.widget_font_size,
                    text=get_hex_code(self.current_color),
                    text_color=BLACK,
                )

    def draw_color_picker(self):
        box_size = int(60 * self.scale)
        spacing_between = int(5 * self.scale)
        step = box_size + spacing_between
        font_size = int(20 * self.scale)
        self.draw_logo_and_help_text()
        self.draw_cursor()

        if self.picked_colors:
            x_start = self.mouse_reg_x - (1.65 * box_size)
            y_start = self.mouse_reg_y - int(box_size / 2)

            x_co, y_co = x_start, y_start
            for index, color in enumerate(self.picked_colors):
                draw_filled_rounded_rectangle(
                    start_x=x_co,
                    start_y=y_co,
                    width=box_size,
                    height=box_size,
                    color=GRAY_PANEL_COLOR,
                    rounding_radius=6,
                )
                draw_filled_rounded_rectangle(
                    start_x=x_co + 2,
                    start_y=y_co + 2,
                    width=box_size - 4,
                    height=box_size - 4,
                    color=color,
                    rounding_radius=6,
                )

                # draw_rectangle(start_x=x_co, start_y=y_co, width=box_size, height=box_size, color=(0.0, 0.0, 0.0, 0.8))
                if self.add_hexcodes_to_colorbox:
                    draw_text(
                        x_co + 5,
                        y_co + 5,
                        font_size=font_size,
                        text=get_hex_code(color),
                        text_color=WHITE if color[0] < 0.6 else BLACK,
                    )
                x_co -= step

                if index % self.colors_per_row == (self.colors_per_row - 1):
                    x_co = x_start
                    y_co -= step

    def modal(self, context, event):
        scene = context.scene
        event_type = event.type
        if event_type in {"WHEELUPMOUSE", "WHEELDOWNMOUSE", "MIDDLEMOUSE"}:
            return {"PASS_THROUGH"}

        elif event_type in {"RIGHTMOUSE", "ESC", "RET", "NUMPAD_ENTER"}:
            if (
                event_type == "RET" or event_type == "NUMPAD_ENTER"
            ) and self.picked_colors:
                if self.collection_index == -1:
                    # it was launched by hotkey, not from a specific collection
                    # Create a new collection!
                    new_collection = scene.color_box.add()
                    new_collection.name = f"{self.image_name}"
                    self.collection_index = len(scene.color_box) - 1

                if self.row_index == -1:
                    new_pallette_name = f"Picked colors from {self.image_name}"
                    add_palette(
                        scene,
                        new_pallette_name,
                        colors=self.picked_colors,
                        collection_index=self.collection_index,
                        open_collection=True,
                    )
                    self.report({"INFO"}, f"Added palette {new_pallette_name}")
                else:
                    palette = scene.color_box[self.collection_index].palettes[
                        self.row_index
                    ]
                    for index, _color in enumerate(self.picked_colors):
                        new_color = palette.colors.add()
                        new_color.color = _color
                    self.report(
                        {"INFO"}, f"Added new colors to {palette.name}"
                    )

            context.area.tag_redraw()
            bpy.types.SpaceImageEditor.draw_handler_remove(
                self._handle, "WINDOW"
            )
            bpy.context.window.cursor_set("DEFAULT")
            return {"CANCELLED"}

        context.area.tag_redraw()

        mouse_reg_x, mouse_reg_y = event.mouse_region_x, event.mouse_region_y
        uv_x, uv_y = context.region.view2d.region_to_view(
            mouse_reg_x, mouse_reg_y
        )
        img_x, img_y = uv_x * self.img_width, uv_y * self.img_height

        self.mouse_reg_x = mouse_reg_x
        self.mouse_reg_y = mouse_reg_y

        bpy.context.window.cursor_set("NONE")

        if 0.0 <= img_x <= self.img_width and 0.0 <= img_y <= self.img_height:
            self.cursor_on_image = True
        else:
            self.cursor_on_image = False
            self.current_color = None
            return {"PASS_THROUGH"}

        x_pixel = math.floor(img_x)
        y_pixel = math.floor(img_y)
        start_index = (y_pixel * self.img_width + x_pixel) * self.channel_count

        # self.current_color = (0.5, 1.0, 0.8, 1.0)  # use this to test how much speedup is possible by caching
        self.current_color = self.pixels[
            start_index : start_index + self.channel_count
        ]

        if (
            event.type == "LEFTMOUSE"
            and event.value == "PRESS"
            and self.cursor_on_image
        ):
            if len(self.picked_colors) < 32:
                self.picked_colors.append(self.current_color)

            if len(self.picked_colors) >= 32:
                self.report(
                    {"INFO"},
                    "The Color Picker is full! You can save or cancel.",
                )

            return {"PASS_THROUGH"}

        elif event.type == "MOUSEMOVE":
            return {"PASS_THROUGH"}

        return {"RUNNING_MODAL"}

    def set_all_variables(self, context, image):
        """
        Only called at invoke, so everything in here will be calculated only once in a session

        The major speed up is accessing the image data once (size, channels, pixels) and most importantly setting
        self.pixels with a python list of pixels (instead of Blender's Image.pixels type)

        Also added a lot of calculation for the widget scaling dimensions/offset here, so it will not be calculated
        again on every mouse move
        """

        try:
            color_box_path = Path(sys.modules[ADDON_NAME].__file__).parent
            # without the parent, it will take the __init__.py file
            logo_file_path = (
                color_box_path
                / Path("example_images")
                / Path("Color_box_raw_logo.png")
            )
            self.logo_image = bpy.data.images.load(
                str(logo_file_path), check_existing=True
            )
            # image.gl_load()  # to load it in GL memory ?
        except Exception as e:
            self.logo_image = None
            print("The logo image could not be loaded")
            print(e)

        self.img_width, self.img_height = image.size
        self.channel_count = image.channels
        self.pixels = image.pixels[:]
        self.image_name = image.name

        # set the following things based on the current user setting, and/or the scale
        preferences = context.preferences.addons[ADDON_NAME].preferences

        # Scaling will take Blender's UI scale into account, and on top of that the ColorBox palette scale
        self.ui_scale = bpy.context.preferences.view.ui_scale
        extra_scale = self.ui_scale * preferences.color_picker_scale
        self.scale *= extra_scale
        self.colors_per_row = preferences.colors_per_row
        self.box_width *= self.scale
        self.display_live_hexcodes = preferences.display_live_hexcodes
        self.add_hexcodes_to_colorbox = preferences.add_hexcodes_to_colorbox

        # specific for live hexcode cursor
        self.widget_diff = int(self.box_width / 4)
        self.widget_half_box_size = int(self.box_width / 2)
        self.widget_hex_text_height = 20 * self.scale
        self.widget_inner_border_width = 6 * self.scale
        self.widget_outer_border_width = 10 * self.scale
        self.widget_outer_width_diff = 14 * self.scale
        self.widget_half_outer_width_diff = int(
            self.widget_outer_width_diff / 2
        )
        self.widget_font_size = int(self.widget_hex_text_height * 1.15)
        self.widget_offset = int(self.widget_hex_text_height * 0.1)
        self.widget_half_outer_border_width = int(
            self.widget_outer_border_width * 0.7
        )

        self.picked_colors = []

    def invoke(self, context, event):
        if context.space_data.ui_mode != "VIEW":
            self.report({"WARNING"}, "Select the View mode to pick colors.")
            return {"CANCELLED"}
        if context.region.type != "WINDOW":
            self.report(
                {"WARNING"}, "Make sure the cursor is not under the UI panel"
            )
            return {"CANCELLED"}

        for area in bpy.context.screen.areas:
            if area.type == "IMAGE_EDITOR" and area.ui_type == "IMAGE_EDITOR":
                self.image_editor_dimensions = area.width, area.height

        image = context.space_data.image
        if image is None:
            self.report({"WARNING"}, "This works only when there is an image")
            return {"CANCELLED"}

        self.set_all_variables(context, image)
        self._handle = bpy.types.SpaceImageEditor.draw_handler_add(
            self.draw_color_picker, (), "WINDOW", "POST_PIXEL"
        )
        context.window_manager.modal_handler_add(self)
        return {"RUNNING_MODAL"}
