import bpy

from ..utils.move_rows import palette_to_ramp, ramp_to_palette
from ..utils.blender import get_material
from ..utils.color_ramp import (
    ColorBoxMaterial,
    add_node_from_data,
    get_color_ramp_node_dict,
)


class COLORBOX_OT_color_ramp_remove(bpy.types.Operator):
    bl_idname = "color_box.color_ramp_remove"
    bl_label = "Delete color ramp"
    bl_options = {"REGISTER"}

    row_index: bpy.props.IntProperty()
    collection_index: bpy.props.IntProperty()

    def execute(self, context):
        scene = context.scene
        node_name = (
            scene.color_box_ramps[self.collection_index]
            .ramps[self.row_index]
            .name
        )
        ColorBoxMaterial.remove_node(node_name=node_name)
        scene.color_box_ramps[self.collection_index].ramps.remove(
            self.row_index
        )

        self.report({"INFO"}, f"Removed ColorRamp node {node_name}")

        return {"FINISHED"}


def get_active_ramp_collection(context):
    """
    Return the first ramp collection that is showed.
    Or the last collection in the list.
    """
    scene = context.scene
    for ramp_collection in scene.color_box_ramps:
        if ramp_collection.show:
            return ramp_collection

    return scene.color_box_ramps[-1] if scene.color_box_ramps else None


class COLORBOX_OT_color_ramp_save(bpy.types.Operator):
    bl_idname = "color_box.color_ramp_save"
    bl_label = "Add the selected Color Ramp node(s) to the Colorbox"
    bl_options = {"REGISTER"}

    def execute(self, context):
        scene = context.scene
        for color_ramp_node in context.selected_nodes:
            if not isinstance(color_ramp_node, bpy.types.ShaderNodeValToRGB):
                continue

            ramp_data = get_color_ramp_node_dict(
                color_ramp_node=color_ramp_node
            )

            color_ramp_node_name = ColorBoxMaterial.add_node(
                ramp_data=ramp_data
            )

            if not scene.color_box_ramps:
                collection = scene.color_box_ramps.add()
                collection.name = "New Color Ramp Collection"
            else:
                collection = get_active_ramp_collection(context)

            new_color_ramp = collection.ramps.add()
            new_color_ramp.name = color_ramp_node_name
            # new_color_ramp.json_data = json.dumps(ramp_data)
            context.area.tag_redraw()
            self.report({"INFO"}, "Color Ramp has been added to the Colorbox")

        return {"FINISHED"}


class COLORBOX_OT_color_ramp_load(bpy.types.Operator):
    bl_idname = "color_box.color_ramp_load"
    bl_label = "Create a material node from the selected Color Ramp"
    bl_options = {"REGISTER"}

    row_index: bpy.props.IntProperty()
    collection_index: bpy.props.IntProperty()

    def execute(self, context):
        material = get_material(context)

        if not material:
            self.report(
                {"ERROR"},
                "There is no material to add this Color Ramp Node to.",
            )
            return {"FINISHED"}
        elif material.name == ColorBoxMaterial.material_name:
            self.report(
                {"ERROR"},
                f"Material {ColorBoxMaterial.material_name} is only for internal use.",
            )
            return {"FINISHED"}

        ramp_data = context.scene.color_box_ramps[self.collection_index].ramps[
            self.row_index
        ]
        # data_dict = json.loads(ramp_data.json_data)
        data_dict = ColorBoxMaterial.get_node_dict(node_name=ramp_data.name)

        material_name, color_ramp_node_name = add_node_from_data(
            data_dict,
            material,
            cursor_location=context.space_data.cursor_location
            if context.area.type == "NODE_EDITOR"
            else None,
        )
        self.report(
            {"INFO"},
            f"A Color Ramp Node has been added for material {material_name}",
        )

        return {"FINISHED"}


class COLORBOX_OT_color_ramp_add(bpy.types.Operator):
    bl_idname = "color_box.color_ramp_add"
    bl_label = "Add a new Color Ramp to this collection"
    bl_options = {"REGISTER"}

    collection_index: bpy.props.IntProperty()

    def execute(self, context):
        """
        Hijack the palette_to_ramp method with some dummy data
        """

        class DummyColor:
            def __init__(self, color):
                self.color = color

        class DummyPallete:
            def __init__(self, colors):
                self.name = "New Color Ramp"
                self.colors = [DummyColor(color) for color in colors]

        palette_to_ramp(
            palette=DummyPallete(
                colors=[(0.0, 0.0, 0.0, 1.0), ((1.0, 1.0, 1.0, 1.0))]
            ),
            color_ramps=context.scene.color_box_ramps,
            collection_to_index=self.collection_index,
        )

        context.area.tag_redraw()
        return {"FINISHED"}


def duplicate_color_ramp(collection_from, collection_to, row_index):
    color_ramp_node_name = ColorBoxMaterial.add_node(
        ramp_data=ColorBoxMaterial.get_node_dict(
            node_name=collection_from.ramps[row_index].name
        )
    )
    new_ramp = collection_to.ramps.add()
    new_ramp.name = color_ramp_node_name


class COLORBOX_OT_color_ramp_duplicate(bpy.types.Operator):
    bl_idname = "color_box.color_ramp_duplicate"
    bl_label = "Duplicate this Color Ramp"
    bl_options = {"REGISTER"}

    def execute(self, context):
        scene = context.scene
        _type, col_index, row_index = scene.color_box_selected_menu
        collection = scene.color_box_ramps[col_index]
        duplicate_color_ramp(
            collection_from=collection,
            collection_to=collection,
            row_index=row_index,
        )
        collection.ramps.move(len(collection.ramps) - 1, row_index + 1)

        context.area.tag_redraw()
        return {"FINISHED"}


class COLORBOX_OT_color_ramp_convert_to_palette(bpy.types.Operator):
    bl_idname = "color_box.color_ramp_convert_to_color_row"
    bl_label = "Convert this Color Ramp into a Palette"
    bl_description = "Make a Palette for every Color Ramp in this collection"
    bl_options = {"REGISTER"}

    def execute(self, context):
        scene = context.scene
        _type, collection_index, row_index = scene.color_box_selected_menu
        ramp_collection = scene.color_box_ramps[collection_index]

        if row_index == -1:
            new_collection = scene.color_box.add()
            new_collection.name = ramp_collection.name
            new_collection_index = len(scene.color_box) - 1

            for row_index, ramp in enumerate(ramp_collection.ramps):
                ramp_to_palette(
                    context,
                    collection_from=collection_index,
                    row_from=row_index,
                    collection_to=new_collection_index,
                )

        else:
            collection_to = int(scene.color_box_options.selected_palette)

            if collection_to == -1:
                new_collection = scene.color_box.add()
                new_collection.name = ramp_collection.ramps[row_index].name
                collection_to = len(scene.color_box) - 1

            ramp_to_palette(
                context,
                collection_from=collection_index,
                row_from=row_index,
                collection_to=collection_to,
            )

        context.area.tag_redraw()
        return {"FINISHED"}
