from contextvars import Context
import random
import bpy

from ..utils.image import inverse_gamma


class COLORBOX_OT_add_to_color_swapper(bpy.types.Operator):
    """Add to Color Swapper Tooltip"""

    bl_idname = "object.add_to_color_swapper"
    bl_label = "Add to Color Swapper"

    def get_swapper_list(self, context):
        swap_collection = context.scene.color_box_swapper

        # the identifier of the enumerator must be a string, that's why the
        # int to string conversion
        return [
            (str(color_index), f"{color_index}: {curr_color.name}", "")
            for color_index, curr_color in enumerate(
                swap_collection.color_collection
            )
        ]

    color_right_click_menu: bpy.props.EnumProperty(
        name="Add to color swapper",
        items=get_swapper_list,
    )

    @classmethod
    def poll(cls, context):
        return (
            hasattr(context, "button_pointer")
            and hasattr(context.button_pointer, "rna_type")
            and isinstance(
                context.button_pointer.rna_type,
                (bpy.types.NodeSocketColor, bpy.types.ColorRampElement),
            )
            and any(cls.get_swapper_list(None, context))
        )

    def split_full_path(self, full_path):
        """
        Expected full_path looks like this for material, world, ColorRamp element or NodeGroup

        bpy.data.materials["Material"].node_tree.nodes["Principled BSDF"].inputs[0].default_value
              bpy.data.worlds["World"].node_tree.nodes["Background"].inputs[0].default_value
        bpy.data.materials["Material"].node_tree.nodes["ColorRamp"].color_ramp.elements[0].color
        bpy.data.node_groups["NodeGroup"].nodes["RGB"].outputs[0].default_value

        Use rsplit to get attribute names (like color, default value).
        Also to get '.elements[0]' or .inputs[12].
        To get the node name, use rsplit('[') becayse node names can
        (and will have) dots (.).
        Color Ramps need an extra rsplit on the .color_ramp. part
        """
        rna, property_name = full_path.rsplit(
            ".", 1
        )  # property name is something like 'default_value', 'color'
        rna, inputs = rna.rsplit(
            ".", 1
        )  # inputs is something like 'inputs[13]', 'outputs[0]', 'elements[0]'

        try:
            assert inputs.endswith(
                "]"
            ), f"Unexpected value {inputs}, expected something like inputs[3]"
            inputs_name, index_str = inputs[:-1].split("[")

            assert inputs_name in {
                "inputs",
                "outputs",
                "elements",
            }, f"This is only implemented for attribute names 'inputs', \
'outputs', 'elements', not for '{inputs_name}'"

            socket_index = int(index_str)
        except Exception as e:
            return None, None, None, None, None

        if (
            inputs_name == "elements"
        ):  # for color ramp elements, there is 1 extra attribute in the path: strip it off
            rna, _ = rna.rsplit(".", 1)

        # get the node name, it is always  .node["Node Name.001"]
        rna, node_name = rna.rsplit("[", 1)
        node_name = node_name[1:-2]

        path, _nodes = rna.rsplit(
            ".", 1
        )  # bpy.data.materials["Material"].node_tree

        if path.endswith(".node_tree"):
            # material_path will be bpy.data.materials["Material"] or bpy.data.worlds["World"]
            material_path, _ = path.rsplit(".", 1)
            # this is the case for most Nodes, including ColorRamps
            assert (
                "ColorBox_ColorRamps" not in material_path
            ), "Cannot add ColorBox ColorRamps to the Swapper, use a Color Ramp added to another material instead"
            material = eval(material_path)
        else:
            # path is most likely bpy.data.node_groups["NodeGroup"] where the name will be NodeGroup.001, etc
            assert (
                ".node_groups" in path
            ), f"Expected '{path}' to end with 'node_tree' or contain 'node_groups'"
            _, node_group_name = path.rsplit(
                '["', 1
            )  # name will include "], like 'NodeGroup"]'
            material = node_group_name[:-2]

        return material, inputs_name, socket_index, node_name, property_name

    def execute(self, context):
        # TODO: if there was something previous on the clipboard,
        # save it and add it back ?

        # convert back the indexing from string to int
        selected_color_index = int(self.color_right_click_menu)

        swap_collection = context.scene.color_box_swapper
        if not (
            0
            <= swap_collection.active_index
            < len(swap_collection.color_collection)
        ):
            self.report(
                {"INFO"},
                f"Make sure there is a Color added  and selected in the Color Swapper",
            )
            return {"FINISHED"}  # CANCELLED

        bpy.ops.ui.copy_data_path_button(full_path=True)
        full_path = context.window_manager.clipboard

        (
            material,
            inputs_outputs,
            socket_index,
            node_name,
            property_name,
        ) = self.split_full_path(full_path)

        new_color_slot_data = {
            "node_name": node_name,  # context.node.name, but will not work for world,
            "socket_index": socket_index,
            # 'pointer_id': str(eval(rna).as_pointer()),  # for future use ? Or determine duplicate by pointer only
            "property_name": property_name,
            "inputs_outputs": inputs_outputs,
        }
        if isinstance(material, str):  # it is a NodeGroup name
            new_color_slot_data["node_group_name"] = material
        elif isinstance(material, bpy.types.Material):
            new_color_slot_data["material"] = material  # context.material
        elif isinstance(material, bpy.types.World):
            new_color_slot_data["world_material"] = material
        elif isinstance(material, bpy.types.Light):
            new_color_slot_data["light_material"] = material

        for _collection_index, _swap_collection in enumerate(
            swap_collection.color_collection
        ):
            for _color_slot_index, _color_slot in enumerate(
                _swap_collection.color_pointers
            ):
                if all(
                    [
                        getattr(_color_slot, _attr_name) == _value
                        for _attr_name, _value in new_color_slot_data.items()
                    ]
                ):
                    # if _collection_index == swap_collection.active_index:
                    if _collection_index == selected_color_index:
                        self.report(
                            {"INFO"},
                            f"This color field was already added for this color slot, "
                            f"at position {_color_slot_index}",
                        )
                        return {"CANCELLED"}
                    else:
                        swap_collection.color_collection[
                            _collection_index
                        ].color_pointers.remove(_color_slot_index)
                        self.report(
                            {"INFO"},
                            f"Reassigned this field. Removed this field from "
                            f"{_collection_index}:{_swap_collection.name}",
                        )  # at position {_color_slot_index}'
                        # not adding break here, will remove one or more occurrences in other places..

        new_color_slot = swap_collection.color_collection[
            selected_color_index
        ].color_pointers.add()

        for attr_name, value in new_color_slot_data.items():
            setattr(new_color_slot, attr_name, value)

        # set the collection color to the selected color
        collection_color_value = swap_collection.color_collection[
            selected_color_index
        ].color
        setattr(
            new_color_slot.resolved_socket,
            new_color_slot.property_name,
            collection_color_value,
        )
        context.area.tag_redraw()

        return {"FINISHED"}


# This class name should not be changed otherwise it will not be picked up for the right context menu (?)
class WM_MT_button_context(bpy.types.Menu):
    bl_label = "WM_MT_button context"

    def draw(self, context):
        pass


def add_color_slot(swap_collection, initial_name="Color"):
    # Clear name collection
    # custom.color_collection.clear()
    new_color_slot = swap_collection.color_collection.add()
    new_color_slot.name = (
        f"{initial_name}{len(swap_collection.color_collection)}"
    )

    # Set active index of both lists to the last item
    swap_collection.active_index = len(swap_collection.color_collection) - 1
    new_color_slot.active_index = len(new_color_slot.color_pointers) - 1


class COLORBOX_OT_swap_collection_add(bpy.types.Operator):
    """ Color fields from your scene can be added to this color """

    bl_idname = "colorbox.swap_collection_add"
    bl_label = "Add a swap color"

    def execute(self, context):
        add_color_slot(context.scene.color_box_swapper)
        return {"FINISHED"}


class COLORBOX_OT_clear_all_swappers(bpy.types.Operator):
    """ Operator to remove all added swappers """

    bl_idname = "colorbox.clear_all_swappers"
    bl_label = "Remove all color swappers"

    def invoke(self, context, event):
        return context.window_manager.invoke_confirm(self, event)

    def execute(self, context):
        context.scene.color_box_swapper.color_collection.clear()
        return {"FINISHED"}


class COLORBOX_OT_swap_collection_add_multiple(bpy.types.Operator):
    """Add multiple color fields"""

    bl_idname = "colorbox.swap_collection_add_multiple"
    bl_label = "Add multiple swappers"

    amount: bpy.props.IntProperty(
        name="Amount of swappers to add", default=2, min=1
    )

    def execute(self, context):
        color_box_swapper = context.scene.color_box_swapper
        for _ in range(self.amount):
            add_color_slot(color_box_swapper, initial_name="Preset")
        return {"FINISHED"}


class COLORBOX_OT_swap_collection_delete(bpy.types.Operator):
    """
    Remove the selected color
    """

    bl_idname = "colorbox.swap_collection_delete"
    bl_label = "Remove this color"

    def execute(self, context):
        swap_collection = context.scene.color_box_swapper
        swap_collection.color_collection.remove(swap_collection.active_index)

        # Set active index of both lists to the last item
        swap_collection.active_index = (
            len(swap_collection.color_collection) - 1
        )
        return {"FINISHED"}


class COLORBOX_OT_swap_color_pointer_delete(bpy.types.Operator):
    """Decouple this color field"""

    bl_idname = "colorbox.swap_color_pointer_delete"
    bl_label = "Remove this color field from the Color swapper"

    def execute(self, context):
        swap_collection = context.scene.color_box_swapper
        color_collection = swap_collection.color_collection[
            swap_collection.active_index
        ]

        color_collection.color_pointers.remove(color_collection.active_index)

        # Set active index of both lists to the last item
        color_collection.active_index = (
            len(color_collection.color_pointers) - 1
        )
        return {"FINISHED"}


def get_collection_and_palette(scene):
    swapper = scene.color_box_swapper
    selected_collection = swapper.selected_color_collection
    selected_palette = swapper.selected_palette

    color_box = scene.color_box
    collection, palette = None, None

    try:
        collection = color_box[selected_collection]
    except IndexError as e:
        return collection, palette

    try:
        palette = collection.palettes[selected_palette]
    except IndexError as e:
        return collection, palette

    return collection, palette


def sync_color_slots(scene):
    """
    When the selected indices for collection and palette can be resolved to an actual collection / palette:
    do nothing.
    Otherwise, set the index or indices to 0.

    This is now used as an easy hack, to make sure that the swapper indices will be set to 0 whenever palettes /
    collections are deleted, or are not present at all.
    This complete flow has to be rewritten, but currently, this Should solve many issues:
    Whenever a new collection or palette will be added or removed, the swapper indices will be 0, and thus cannot
    continue to point to a larger non existing collection.
    """
    collection, palette = get_collection_and_palette(scene)
    swapper = scene.color_box_swapper

    if collection is None:
        swapper.selected_color_collection = 0
        swapper.selected_palette = 0
        return

    if palette is None:
        swapper.selected_palette = 0


def apply_palette(swapper, palette):
    color_slots = swapper.color_collection
    palette_colors = palette.colors[:16]
    slots_to_add = max(0, len(palette_colors) - len(color_slots)) if swapper.auto_add_swaps else 0
    slots_total = len(color_slots) + slots_to_add

    for i in range(slots_to_add):
        add_color_slot(swapper)

    for color_index, weighted_color in enumerate(palette_colors[:slots_total]):
        color = weighted_color.color
        color_slots[color_index].color = (
            tuple(
                inverse_gamma(
                    r,
                )
                for r in color[:3]
            )
            + (color[3],)
        )


class COLORBOX_OT_swap_colorbox_navigation(bpy.types.Operator):
    """ Browse through this palette or collection """

    bl_idname = "colorbox.swap_colorbox_navigation"
    bl_label = "Navigate"
    bl_options = {'REGISTER', 'UNDO', 'INTERNAL'}

    navigation_type: bpy.props.IntVectorProperty(size=2, default=(-1, 0))

    def execute(self, context):
        # TODO: clean up this mess
        scene = context.scene
        navigation_type, addition = self.navigation_type

        if addition not in {-1, 1}:
            self.report({"WARNING"}, f"Wrong navigation input")
            return {"CANCELLED"}

        swapper = scene.color_box_swapper
        selected_collection = swapper.selected_color_collection
        selected_palette = swapper.selected_palette

        if navigation_type == 0:
            selected_collection += addition
            selected_palette = 0
        elif navigation_type == 1:
            selected_palette += addition

        # before setting these indices, make sure they are allowed in the color box
        color_box = scene.color_box
        if len(color_box) == 0:
            selected_collection = 0
            selected_palette = 0
        else:
            selected_collection = selected_collection % len(color_box)

            palette_count = len(color_box[selected_collection].palettes)
            if palette_count == 0:
                selected_palette = 0
            else:
                selected_palette = selected_palette % palette_count

        swapper.selected_color_collection = selected_collection
        swapper.selected_palette = selected_palette

        if swapper.auto_apply:
            collection, palette = get_collection_and_palette(scene)
            if collection is not None and palette is not None:
                apply_palette(swapper, palette)

        context.area.tag_redraw()
        return {"FINISHED"}


class COLORBOX_OT_swap_apply_palette(bpy.types.Operator):
    """ Apply this palette (in order) to the color slots below """

    bl_idname = "colorbox.swap_apply_palette"
    bl_label = "Apply palette"

    def execute(self, context):
        scene = context.scene
        collection, palette = get_collection_and_palette(scene)

        if collection is None or palette is None:
            self.report({"WARNING"}, f"Could not apply this palette")
            return {"CANCELLED"}

        swapper = scene.color_box_swapper
        apply_palette(swapper, palette)

        return {"FINISHED"}


class COLORBOX_OT_swap_shuffle(bpy.types.Operator):
    """ Shuffle the colors in the Color Swapper """

    bl_idname = "colorbox.swap_shuffle"
    bl_label = "Shuffle colors"

    def execute(self, context):
        color_collection = context.scene.color_box_swapper.color_collection

        shuffled_colors = [tuple(c.color) for c in color_collection]
        random.shuffle(shuffled_colors)

        for index, color in enumerate(shuffled_colors):
            color_collection[index].color = color

        context.area.tag_redraw()
        return {"FINISHED"}


class COLORBOX_OT_create_swapper_nodes(bpy.types.Operator):
    bl_idname = "color_box.create_swapper_nodes"
    bl_label = "Create Nodes"
    bl_options = {"REGISTER"}

    def execute(self, context: Context):
        swap_collection = context.scene.color_box_swapper.color_collection

        curr_material = context.active_object.active_material
        node_tree = curr_material.node_tree

        group = bpy.data.node_groups.new(
            type="ShaderNodeTree", name="COLORBOX"
        )
        nodes = group.nodes
        output_node = group.nodes.new("NodeGroupOutput")
        output_node.location = (300, 0)

        for col_index, curr_color in enumerate(swap_collection):
            color = curr_color.color
            created_node = nodes.new("ShaderNodeRGB")
            created_node.location = (0, col_index * -200)

            socket_out_name = f"Out_{col_index}"

            try:
                group.interface.new_socket(socket_type="NodeSocketColor", name = socket_out_name,in_out='OUTPUT')
            except RuntimeError:
                print("Running a Blender version that doesn't \
                       need to a create socket")

            node_output = created_node.outputs[0]
            node_output.default_value = color
            # print(color[0],color[1],color[2],color[3])

            group.links.new(node_output, output_node.inputs[-2])

            color_slot = curr_color.color_pointers.add()
            color_slot.node_name = created_node.name
            color_slot.node_group_name = group.name
            color_slot.material = curr_material
            color_slot.inputs_outputs = "outputs"
            color_slot.property_name = "default_value"
            color_slot.index = 0


        group_node = node_tree.nodes.new("ShaderNodeGroup")
        group_node.node_tree = group

        return {"FINISHED"}
