from pathlib import Path

import bpy

from ..utils.move_rows import add_palette
from ..utils.image import PixelIterator, round_color_tuple
from .. import ADDON_ID

PACKAGE_NAME = ADDON_ID
CB_PNG_COLOR_WIDTH = 50
CB_PNG_COLOR_HEIGHT = 50
CB_PNG_FILENAME_PREFIX = "CB_"


def clean_CB_prefix_from_filename(name):
    """
    If the file was preficed with 01CB_somename, it wil be returned as somename.
    If the first part does not match exactly with 'some integers followed by exactly "CB_"',
    the original name will be returned
    """
    cb_match = name.find(CB_PNG_FILENAME_PREFIX)

    if cb_match == -1:
        return name

    try:
        # the first parth has to be something like 00, 01, 007 that can be converted into an integer
        int(name[:cb_match])
    except Exception as e:
        # the first part contained other character than integers only, it was not generated by colorbox
        return name

    return name[cb_match + len(CB_PNG_FILENAME_PREFIX) :]


def make_png_from_palette(palette, output_path: str):
    """
    Add all colors from this palette in one row.
    This means that the pixel of the first color has to appear CB_PNG_COLOR_WIDTH times in a row,
    and then followed by the second color and so on.
    After the first pixel row is done, that list can me multiplied by COLOR_HEIGHT, to give the output image some height
    """
    colors = [weighted_color.color for weighted_color in palette.colors]

    pixels = []
    for color in colors:
        pixels += list(color) * CB_PNG_COLOR_WIDTH
    pixels *= CB_PNG_COLOR_HEIGHT

    image_width = len(colors) * CB_PNG_COLOR_WIDTH
    image = bpy.data.images.new(
        palette.name, width=image_width, height=CB_PNG_COLOR_HEIGHT
    )
    image.pixels = pixels
    image.filepath_raw = output_path
    image.file_format = "PNG"
    image.save()


class BaseExportPng(bpy.types.Operator):
    bl_label = "Generate a png file from this color row or this collection."
    bl_options = {"REGISTER"}

    def execute(self, context):
        (
            _type,
            collection_index,
            row_index,
        ) = context.scene.color_box_selected_menu

        if not (collection_index >= 0 and _type == 0):
            self.report({"ERROR"}, "Could not generate an png")
            return {"FINISHED"}

        addon_prefs = context.preferences.addons[PACKAGE_NAME].preferences
        collection = context.scene.color_box[collection_index]

        # when the user selected a specific path, it is not the empty string.
        # when it is the empty string, it will be resolved to the home folder, but we try to grab Desktop.
        # only set it if it exists
        png_path = addon_prefs.export_png_folder
        output_dir = Path(png_path)
        if png_path == "":
            desktop_path = output_dir / Path("Desktop")
            if desktop_path.exists():
                output_dir = desktop_path

        if row_index == -1:
            all_palettes = collection.palettes

            new_dir_name = f"{collection.name.replace(' ', '_')}_CB"
            output_dir /= Path(new_dir_name)
            output_dir.mkdir(parents=True, exist_ok=True)

            max_index_length = len(str(len(all_palettes))) + 1
            for pal_index, palette in enumerate(all_palettes):
                prefix = f'{str(pal_index).rjust(max_index_length, "0")}{CB_PNG_FILENAME_PREFIX}'
                make_png_from_palette(
                    palette,
                    output_path=str(
                        output_dir / Path(f"{prefix}{palette.name}.png")
                    ),
                )

            self.report({"INFO"}, f"Save png(s) in directory {output_dir}")
        else:
            palette = collection.palettes[row_index]
            file_name = Path(f"{CB_PNG_FILENAME_PREFIX}{palette.name}.png")
            output_path = str(output_dir / file_name)
            make_png_from_palette(palette, output_path=output_path)
            self.report({"INFO"}, f"Saved {output_path}")

        return {"FINISHED"}


class COLORBOX_OT_export_png(BaseExportPng):
    bl_idname = "color_box.export_png"
    bl_description = "Generate a .png file from this color row."


class COLORBOX_OT_export_png_collection(BaseExportPng):
    bl_idname = "color_box.export_png_collection"
    bl_description = "Create a folder with the collection name, containing .png files for every row in this collection."


def is_color_similar(color1, color2, max_diff=0.05):
    """
    Return as early as possible, False whenever rgb values (color[0], color[1] or color[2])
    of the two colors differ more then the max.
    """
    for i in range(3):
        if abs(color1[i] - color2[i]) > max_diff:
            return False

    return True


def get_distinct_colors(color_tuples, maximum_colors, max_color_diff=0.05):
    distinct_colors = []

    for _count, color in color_tuples:
        new_color = True
        for dist_color in distinct_colors:
            if is_color_similar(dist_color, color, max_diff=max_color_diff):
                new_color = False
                break

        if new_color:
            distinct_colors.append(color)

        if len(distinct_colors) == maximum_colors:
            break

    return distinct_colors


def extract_from_image(image, preferences):
    """
    Use the user preferences to determine colors
    """
    max_colors_to_return = preferences.max_colors_to_return
    determine_distinct_colors = preferences.determine_distinct_colors
    pixel_threshold = preferences.pixel_threshold

    new_colors = {}
    for new_color in PixelIterator(
        image=image, analyse_all_pixels=preferences.analyse_all_pixels
    ):
        if new_color[3] >= preferences.ignore_alpha_below:
            new_colors.setdefault(new_color, 0)
            new_colors[new_color] += 1

    color_tuples = sorted(
        [(count, color) for color, count in new_colors.items()], reverse=True
    )
    if pixel_threshold > 0.0:
        highest_frequency = color_tuples[0][0]
        threshold = int(highest_frequency / pixel_threshold)
        color_tuples = [
            (count, color)
            for count, color in color_tuples
            if count > threshold
        ]

    colors = []
    weights = []
    if preferences.use_weighted_colors:
        total_counts = sum([count for count, _ in color_tuples])
        for count, color in color_tuples:
            colors.append(color)
            weights.append(count / total_counts)
        colors = colors[::-1]
    else:
        if determine_distinct_colors > 0:
            colors = get_distinct_colors(
                color_tuples, max_colors_to_return, determine_distinct_colors
            )
        else:
            colors = [color for _, color in color_tuples]

    return colors[:max_colors_to_return], weights[:max_colors_to_return], None


def extract_adobe_colors(image):
    channel_count = image.channels
    pixels = image.pixels[:]  # this takes 99% of the time

    def get_color(x, y):
        start_index = (y * 2480 + x) * channel_count
        return round_color_tuple(
            pixels[start_index : start_index + channel_count]
        )

    colors = []
    x_offset = 248  # do not take the border
    for color_index in range(5):
        samples = [get_color(x_offset, y) for y in range(300, 1000, 100)]

        assert len(set(samples)) == 1, "This is not an Adobe image"

        colors.append(samples[0])
        x_offset += 496  # this is 1/5th of the image width

    return colors, [], "ADOBE"


def extract_colorbox_colors(image):
    channel_count = image.channels
    pixels = image.pixels[:]  # this takes 99% of the time

    start_pixel = int(CB_PNG_COLOR_WIDTH / 2)
    len_colors = image.size[0] // CB_PNG_COLOR_WIDTH

    def get_color(x):
        start_index = x * channel_count
        return round_color_tuple(
            pixels[start_index : start_index + channel_count]
        )

    colors = [
        get_color(x=start_pixel + i * CB_PNG_COLOR_WIDTH)
        for i in range(len_colors)
    ]
    return colors, [], "Colorbox"


def extract_from_image_palette(image):
    """
    Use image dimensions to guess the source type, then return colors based on that.
    """
    max_colors_to_return = 12

    image_width, image_height = image.size
    if (
        tuple(image.size) in {(1335, 668), (400, 400)}
        and image.file_format == "PNG"
    ):
        # COLOR HUNT, new format is 400x400 pixels
        max_colors_to_return = 4
    elif tuple(image.size) == (2480, 1040):  # ADOBE
        return extract_adobe_colors(image)
    elif (
        image_height == CB_PNG_COLOR_HEIGHT
        and image_width % CB_PNG_COLOR_WIDTH == 0
        and image.file_format == "PNG"
    ):
        return extract_colorbox_colors(
            image
        )  # very likely to be generated by Colorbox
    elif image_width <= 10 and image_height <= 10:
        max_colors_to_return = 100  # only for very small images, we expect a pixel palette with up to 100 colors

    if image_width * image_height > 100000:
        pixel_iterator = PixelIterator(image=image, analyse_all_pixels=False)
    else:
        pixel_iterator = PixelIterator(image=image, analyse_all_pixels=True)

    new_colors = {}
    for new_color in pixel_iterator:
        if new_color[3] >= 1.00:
            new_colors.setdefault(new_color, 0)
            new_colors[new_color] += 1

    color_tuples = sorted(
        [(count, color) for color, count in new_colors.items()], reverse=True
    )
    color_tuples = color_tuples[:max_colors_to_return]

    if max_colors_to_return == 4:  # Only for ColorHunt, collect weights
        colors = []
        weights = []
        total_counts = sum([count for count, _ in color_tuples])
        for count, color in color_tuples:
            colors.append(color)
            weights.append(count / total_counts)
        colors = colors[::-1]

        return colors, weights, None

    # if there are more than 6 colors, calculate a threshold, to eliminate very infrequent colors
    # take the average of the first 4 most frequent colors (if there are 4), to use that value to determine
    # infrequent colors
    first_colors = [count for count, _ in color_tuples[:4]]
    if first_colors:
        frequency = sum(first_colors) / len(first_colors)
        threshold = int(frequency / 60)
        colors = [color for count, color in color_tuples if count > threshold]
    else:
        colors = [color for _, color in color_tuples]

    return colors[:max_colors_to_return], [], None


class BaseImageExtractionOperator(bpy.types.Operator):
    bl_idname = "color_box.extract_from_image"
    bl_label = "Extract colors"
    bl_description = "Upload file or folder"

    filepath: bpy.props.StringProperty(
        name="File Path",
        description="Filepath used for importing the file",
    )
    directory: bpy.props.StringProperty(
        name="Directory",
        description="Extract colors",
    )

    bl_options = {"REGISTER"}

    def _return(self, error_msg=None, info_msg=None):
        if error_msg:
            self.report({"ERROR"}, error_msg)
        elif info_msg:
            self.report({"INFO"}, info_msg)

        return {"FINISHED"}

    def get_collection_index(self, context, path):
        """
        Create a new collection, with either the directory name or the single file name.
        Then return the index for that collection
        """
        new_collection = context.scene.color_box.add()
        new_collection.name = path.stem
        new_collection.show = True
        return len(context.scene.color_box) - 1

    def extract_from_image(self, context, image_path):
        image = bpy.data.images.load(image_path, check_existing=False)
        channel_count = image.channels

        if image.file_format not in ["JPEG", "PNG"]:
            return self._return(
                error_msg=f"Currently, this only works for JPEG and PNG image files"
            )
        if channel_count not in [3, 4]:
            return self._return(
                error_msg=f"This image has {channel_count} channels, but this method can only handle 3 or 4 channels"
            )

        if self.strategy == "custom":
            return extract_from_image(
                image,
                preferences=context.preferences.addons[
                    PACKAGE_NAME
                ].preferences,
            )
        else:
            return extract_from_image_palette(image)

    def _execute_single_image(self, context, image_path):
        scene = context.scene
        pathlib_path = Path(image_path)

        # early check on the path, to not load unnecessary things in Blender
        if not image_path or pathlib_path.suffix.lower() not in [
            ".jpeg",
            ".png",
        ]:
            return self._return(error_msg=f"Please select a JPEG or PNG file.")
        try:
            weighted_colors = self.extract_from_image(
                context, image_path=image_path
            )
            if weighted_colors == {"FINISHED"}:  # error was already reported
                return self._return()

            colors, weights, _palette_type = weighted_colors
            palette_name = pathlib_path.stem
            if _palette_type == "ADOBE" and palette_name.startswith(
                "AdobeColor-"
            ):
                palette_name = palette_name[11:]
            if _palette_type == "Colorbox":
                palette_name = clean_CB_prefix_from_filename(palette_name)

            add_palette(
                scene,
                palette_name=palette_name,
                colors=colors,
                weights=weights,
                collection_index=self.get_collection_index(
                    context, pathlib_path
                ),
            )

            context.area.tag_redraw()
            return self._return(
                info_msg=f"Colorbox row saved for {pathlib_path}"
            )

        except Exception as e:
            print(repr(e))
            return self._return(
                error_msg=f"Could not extract colors from '{pathlib_path}'"
            )

    def _execute_dir(self, context, dir_path):
        scene = context.scene
        pathlib_path = Path(dir_path)
        if not dir_path or not pathlib_path.is_dir():
            return self._return(error_msg=f"Please select a directory.")

        # create a new collection with the name of the directory
        collection_index = self.get_collection_index(context, pathlib_path)

        palettes_added = 0
        errors = 0
        for image_path in sorted(pathlib_path.iterdir()):
            if image_path.is_file() and image_path.suffix.lower() in [
                ".jpeg",
                ".png",
            ]:
                try:
                    weighted_colors = self.extract_from_image(
                        context, image_path=str(image_path)
                    )
                    if weighted_colors == {
                        "FINISHED"
                    }:  # the error was reported, just continue with the next image
                        errors += 1
                        continue

                    colors, weights, _palette_type = weighted_colors

                    palette_name = image_path.stem
                    if _palette_type == "ADOBE" and palette_name.startswith(
                        "AdobeColor-"
                    ):
                        palette_name = palette_name[11:]
                    if _palette_type == "Colorbox":
                        palette_name = clean_CB_prefix_from_filename(
                            palette_name
                        )

                    add_palette(
                        scene,
                        palette_name=palette_name,
                        colors=colors,
                        collection_index=collection_index,
                        weights=weights,
                    )
                    palettes_added += 1
                except Exception as e:
                    print(
                        f"Exception when trying to extract a palette from {image_path}"
                    )
                    print(repr(e))
                    errors += 1

        if palettes_added == 0 and errors:
            scene.color_box.remove(collection_index)
            self.report(
                {"ERROR"},
                f"Error while extracting colors from '{pathlib_path}'. Does the folder contain JPEG or PNG files?",
            )
        else:
            error_part = (
                f"{errors} files were not processed." if errors else ""
            )
            self.report(
                {"INFO"}, f"{palettes_added} Colorbox rows saved. {error_part}"
            )

        context.area.tag_redraw()
        return {"FINISHED"}

    def clean_file_path(self):
        """
        context.window_manager.fileselect_add(self) (called from invoke), will do some things with self.filepath
        beyond my understanding. (it will be set to the current blend file, in case this is called from a saved file.

        In any case when the filepath ends with .blend, strip it from the path.

        This solves the issue where a user chose a directory (and then the fileselect_add adds the current .blend to
        the path).
        The only unexpected behaviour would happen when a use purposely selects a .blend file. This does not make any
        sense, but the code will act as if the directory was selected.
        """
        if self.filepath.endswith(".blend"):
            self.filepath = str(Path(*Path(self.filepath).parts[:-1]))

    def execute(self, context):
        self.clean_file_path()
        file_path = self.filepath
        # in any case, reset the filepath, if not, it cannot be determined if the next selection is a folder or not
        self.filepath = ""

        if Path(file_path) == Path(self.directory) and file_path != "":
            return self._execute_dir(context, dir_path=self.directory)

        elif file_path != "":
            return self._execute_single_image(context, image_path=file_path)

        else:
            return self._return(error_msg="Something went wrong")

    def invoke(self, context, _event):
        context.window_manager.fileselect_add(self)
        return {"RUNNING_MODAL"}


class COLORBOX_OT_extract_from_palette(BaseImageExtractionOperator):
    bl_idname = "color_box.extract_from_palette"
    bl_label = "Import palettes"
    bl_description = "Upload file or folder with palettes"

    strategy = "palette"


class COLORBOX_OT_extract_from_image(BaseImageExtractionOperator):
    bl_idname = "color_box.extract_from_image"
    bl_label = "Custom import"
    bl_description = "Upload file or folder with any image"

    strategy = "custom"
