import bpy
from colorsys import rgb_to_hsv, hsv_to_rgb
import re

from bpy.types import Context

from ..utils.move_rows import palette_to_ramp, add_palette
from ..utils.image import (
    inverse_gamma,
    correct_gamma,
    round_color_tuple,
    get_hex_code,
)
from ..utils.import_export import (
    import_color_box,
    get_color_box_data_path,
    export_color_box,
)
from .. import ADDON_ID

PACKAGE_NAME = ADDON_ID


class COLORBOX_OT_load_json(bpy.types.Operator):
    """
    It is located at config/color_box.json
    """

    bl_idname = "color_box.load_json"
    bl_label = "Load the Colorbox from the config"
    bl_options = {"REGISTER"}

    def execute(self, context):
        color_box_data_path = get_color_box_data_path()

        if not color_box_data_path.exists():
            self.report(
                {"INFO"},
                f"No Colorbox has been saved at {color_box_data_path}",
            )
            return {"FINISHED"}

        try:
            import_color_box(color_box_data_path, context=context)
            self.report(
                {"INFO"}, f"Colorbox loaded from {color_box_data_path}"
            )
        except Exception as e:
            self.report(
                {"ERROR"},
                f"Something went wrong while loading {color_box_data_path}. "
                f"Check the error message in the Python console.",
            )
            print(repr(e))

        context.area.tag_redraw()
        return {"FINISHED"}


class COLORBOX_OT_next_palette_page(bpy.types.Operator):
    bl_idname = "color_box.next_palette_page"
    bl_label = "Next Palette Page"
    bl_options = {"REGISTER"}

    @classmethod
    def poll(cls, context):
        scene = context.scene
        number_of_pages = scene.color_box_palette_page_amount - 2

        if scene.color_box_total_palettes > scene.color_box_palette_per_page:
            number_of_pages += 1

        return scene.color_box_palette_current_page < number_of_pages

    def execute(self, context):
        context.scene.color_box_palette_current_page += 1
        return {"FINISHED"}


class COLORBOX_OT_previous_palette_page(bpy.types.Operator):
    bl_idname = "color_box.previous_palette_page"
    bl_label = "Previous Palette Page"
    bl_options = {"REGISTER"}

    @classmethod
    def poll(cls, context):
        scene = context.scene
        return scene.color_box_palette_current_page > 0

    def execute(self, context):
        context.scene.color_box_palette_current_page -= 1
        return {"FINISHED"}


class COLORBOX_OT_save_json(bpy.types.Operator):
    """
    Save all current palettes and color ramps to config/color_box.json
    """

    bl_idname = "color_box.save_json"
    bl_label = "Save the Colorbox from the scene to the color_box.json"
    bl_options = {"REGISTER"}

    def execute(self, context):
        color_box_data_path = get_color_box_data_path()
        export_color_box(data_path=color_box_data_path, context=context)
        self.report({"INFO"}, f"Colorbox saved as {color_box_data_path}")

        return {"FINISHED"}


# class COLORBOX_OT_load_default(bpy.types.Operator):
#     """
#     This one has become obsolete now, could be deleted, unless we decide to use it anyway.
#     Load the default colors. Clear the current Colorbox first.
#
#     Right now, the defaults are hardcoded, so there is an example to show in the interface.
#     This will be replaced by a loading the default from a file that a user saved on disc (via a ColorBox operator)
#     """
#     bl_idname = 'color_box.load_default'
#     bl_label = 'Load default Colorbox'
#     bl_options = {'REGISTER'}
#
#     def execute(self, context):
#         context.scene.color_box.clear()
#
#         add_palette(context, palette_name='Pinks', colors=[(0.90, 0.10, 0.29), (0.75, 0.16, 0.37), (0.60, 0.23, 0.45)]
#         add_palette(
#             context, palette_name='Ocean',
#             colors=[(0.0, 0.51, 0.78), (0.16, 0.57, 0.67), (0.33, 0.63, 0.55), (0.50, 0.69, 0.44)],
#             weights=[0.4, 0.3, 0.15, 0.15])
#         add_palette(context, palette_name='Sunset', colors=[
#             (0.996, 0.88, 0.098),
#             (0.98, 0.75, 0.13),
#             (0.96, 0.62, 0.16),
#             (0.95, 0.49, 0.20),
#             (0.93, 0.36, 0.23),
#             (0.92, 0.23, 0.26)
#         ])
#
#         return {'FINISHED'}


class COLORBOX_OT_add_row(bpy.types.Operator):
    bl_idname = "color_box.add_color_row"
    bl_label = "Add a new color row"
    bl_options = {"REGISTER"}

    collection_index: bpy.props.IntProperty()

    def execute(self, context):
        from .color_swapper import sync_color_slots

        scene = context.scene

        add_palette(
            scene,
            palette_name="New row",
            colors=[(1, 1, 1, 1)],
            collection_index=self.collection_index,
        )
        sync_color_slots(scene)
        context.area.tag_redraw()
        return {"FINISHED"}


class COLORBOX_OT_remove_row(bpy.types.Operator):
    bl_idname = "color_box.remove_color_row"
    bl_label = "Delete palette"
    bl_options = {"REGISTER"}

    collection_index: bpy.props.IntProperty()
    row_index: bpy.props.IntProperty()

    def execute(self, context):
        # TODO: Perhaps move imports to the top of the module
        from ..utils.move_rows import update_pagination
        from .color_swapper import sync_color_slots

        collection_index = self.collection_index
        scene = context.scene

        palettes = scene.color_box[collection_index].palettes
        palettes.remove(self.row_index)

        update_pagination(scene, palettes)
        sync_color_slots(scene)

        context.area.tag_redraw()
        return {"FINISHED"}


def check_color_value(value):
    if value < 0:
        return check_color_value(abs(value))
    elif value > 1:
        return max(0, 1 - (value - 1))

    return value


class COLORBOX_OT_add_color(bpy.types.Operator):
    bl_idname = "color_box.add_color"
    bl_label = "Add a new color"
    bl_options = {"REGISTER"}

    collection_index: bpy.props.IntProperty()
    row_index: bpy.props.IntProperty()

    def execute(self, context):
        palette = context.scene.color_box[self.collection_index].palettes[
            self.row_index
        ]

        new_color = [1, 1, 1, 1]  # default value will be 0

        if len(palette.colors) == 1:
            last_color = palette.colors[-1].color
            addition = 0.18
            if (
                last_color[0] > 0.9
                or last_color[1] > 0.9
                or last_color[2] > 0.9
            ):
                addition *= -1
            new_color = [last_color[i] + addition for i in range(3)]

        elif len(palette.colors) > 1:
            last_color = palette.colors[-1].color
            second_last_color = palette.colors[-2].color

            new_color = [
                last_color[i] + last_color[i] - second_last_color[i]
                for i in range(3)
            ]

        if len(new_color) == 3:
            new_color.append(1)

        palette.colors.add().color = [check_color_value(v) for v in new_color]

        context.area.tag_redraw()
        return {"FINISHED"}


class COLORBOX_OT_remove_color(bpy.types.Operator):
    bl_idname = "color_box.remove_color"
    bl_label = "Remove the last color"
    bl_options = {"REGISTER"}

    collection_index: bpy.props.IntProperty()
    row_index: bpy.props.IntProperty()

    def execute(self, context):
        palette = context.scene.color_box[self.collection_index].palettes[
            self.row_index
        ]
        color_count = len(palette.colors)

        if color_count <= 1:
            self.report(
                {"INFO"},
                "You are not allowed to delete the last color. You can change it, or remove the entire row.",
            )
        else:
            # only remove the last item when there are more then one colors. If there is only one color,
            # it cannot be deleted. The user could delete the entire row in that case.
            palette.colors.remove(color_count - 1)

        context.area.tag_redraw()
        return {"FINISHED"}


class COLORBOX_OT_import_blender_palettes(bpy.types.Operator):
    """
    Import all Blender palettes (they can be find in Texture Paint --> Tool --> Color Palette)
    """

    bl_idname = "color_box.import_blender_palettes"
    bl_label = "Import the color palettes as defined in the Texture Paint tab."
    bl_options = {"REGISTER"}

    def execute(self, context):
        scene = context.scene

        for palette in bpy.data.palettes:
            if not palette.colors:
                continue

            new_collection = context.scene.color_box.add()
            new_collection.name = "Blender palettes"
            add_palette(
                scene,
                palette_name=palette.name,
                collection_index=len(context.scene.color_box) - 1,
                colors=[
                    (c.color[0], c.color[1], c.color[2], 1)
                    for c in palette.colors
                ],
            )

        context.area.tag_redraw()
        return {"FINISHED"}


def get_palette(context):
    _type, col_index, row_index = context.scene.color_box_selected_menu

    palette = None
    collection = None
    if _type == 0 and col_index >= 0:
        collection = context.scene.color_box[col_index]

        if row_index >= 0:
            palette = collection.palettes[row_index]

    return collection, palette


def get_palettes(context):
    """
    Return a list of palettes.
    When a specific row_index is present, the list of palettes contains that palette only
    When no specific row index is present (it is -1), palettes is a list of all palettes
    """
    _type, col_index, row_index = context.scene.color_box_selected_menu

    if not (_type == 0 and col_index >= 0):
        return []

    collection = context.scene.color_box[col_index]
    if row_index >= 0:
        return [collection.palettes[row_index]]
    else:
        return collection.palettes


class COLORBOX_OT_reverse_colors(bpy.types.Operator):
    bl_idname = "color_box.reverse_colors"
    bl_label = "Reverse the order of colors"
    bl_options = {"REGISTER"}

    def execute(self, context):
        for palette in get_palettes(context):
            last_index = len(palette.colors) - 1
            for i in range(last_index):
                palette.colors.move(last_index, i)

        context.area.tag_redraw()
        return {"FINISHED"}


class COLORBOX_OT_apply_gamma(bpy.types.Operator):
    bl_idname = "color_box.apply_gamma"
    bl_label = "Apply gamma correction to this color row"
    bl_options = {"REGISTER"}

    def execute(self, context):
        gamma_value = context.preferences.addons[
            PACKAGE_NAME
        ].preferences.gamma_value

        for palette in get_palettes(context):
            for weighted_color in palette.colors:
                color = weighted_color.color
                weighted_color.color = tuple(
                    correct_gamma(r, gamma_value=gamma_value)
                    for r in color[:3]
                ) + (color[3],)

        context.area.tag_redraw()
        return {"FINISHED"}


class COLORBOX_OT_apply_inverse_gamma(bpy.types.Operator):
    bl_idname = "color_box.apply_inverse_gamma"
    bl_label = "Apply inverse gamma correction to this color row"
    bl_options = {"REGISTER"}

    def execute(self, context):
        gamma_value = context.preferences.addons[
            PACKAGE_NAME
        ].preferences.gamma_value

        for palette in get_palettes(context):
            for weighted_color in palette.colors:
                color = weighted_color.color
                weighted_color.color = tuple(
                    inverse_gamma(r, gamma_value=gamma_value)
                    for r in color[:3]
                ) + (color[3],)

        context.area.tag_redraw()
        return {"FINISHED"}


def copy_palette(collection, palette, name_prefix=None):
    new_palette = collection.palettes.add()
    new_name = (
        palette.name if not name_prefix else f"{palette.name}{name_prefix}"
    )
    new_palette.name = new_name

    for weighted_color in palette.colors:
        new_color = new_palette.colors.add()

        new_color.color = weighted_color.color
        new_color.weight = weighted_color.weight


class COLORBOX_OT_create_nodes(bpy.types.Operator):
    bl_idname = "color_box.create_nodes"
    bl_label = "Create Nodes"
    bl_options = {"REGISTER"}

    def execute(self, context: Context):
        collection, palette = get_palette(context)

        curr_material = context.active_object.active_material.node_tree
        group = bpy.data.node_groups.new(
            type="ShaderNodeTree", name=palette.name
        )
        nodes = group.nodes
        output_node = group.nodes.new("NodeGroupOutput")
        output_node.location = (300, 0)

        for col_index, curr_color in enumerate(reversed(palette.colors)):
            color = curr_color.color
            created_node = nodes.new("ShaderNodeRGB")
            created_node.location = (0, col_index * -200)

            socket_out_name = f"Out_{col_index}"

            group.interface.new_socket(socket_type="NodeSocketColor", name =socket_out_name,in_out='OUTPUT')

            node_output = created_node.outputs[0]
            node_output.default_value = color

            group.links.new(node_output, output_node.inputs[-1])

        group_node = curr_material.nodes.new("ShaderNodeGroup")
        group_node.node_tree = group

        return {"FINISHED"}


class COLORBOX_OT_copy_color_row(bpy.types.Operator):
    bl_idname = "color_box.copy_row"
    bl_label = "Duplicate palette"
    bl_options = {"REGISTER"}

    def execute(self, context):
        collection, palette = get_palette(context)

        copy_palette(collection, palette, name_prefix="_copy")

        row_index = context.scene.color_box_selected_menu[2]
        collection.palettes.move(len(collection.palettes) - 1, row_index + 1)

        context.area.tag_redraw()
        return {"FINISHED"}


class COLORBOX_OT_sort_by_hue(bpy.types.Operator):
    bl_idname = "color_box.sort_by_hue"
    bl_label = "Sort this row by hue"
    bl_options = {"REGISTER"}

    def execute(self, context):
        for palette in get_palettes(context):
            original_colors = sorted(
                [
                    (
                        rgb_to_hsv(*wc.color[:3]),
                        tuple(wc.color),
                        wc.weight,
                        wc.selected,
                    )
                    for wc in palette.colors
                ]
            )
            for new_index, (hsv, color, weight, selected) in enumerate(
                original_colors
            ):
                palette.colors[new_index].color = color
                palette.colors[new_index].weight = weight
                palette.colors[new_index].selected = selected

        context.area.tag_redraw()
        return {"FINISHED"}


class COLORBOX_OT_add_to_color_ramp(bpy.types.Operator):
    bl_idname = "color_box.add_to_color_ramp"
    bl_label = "Make a Color Ramp from this palette"
    bl_description = "Make a Color Ramp for all palettes in this collection"
    bl_options = {"REGISTER"}

    def execute(self, context):
        _type, col_index, row_index = context.scene.color_box_selected_menu

        if not (_type == 0 and col_index >= 0):
            return []

        collection = context.scene.color_box[col_index]
        if row_index >= 0:
            palette = collection.palettes[row_index]

            collection_to_index = int(
                context.scene.color_box_options.selected_color_ramp
            )
            if collection_to_index > -1:
                # a specific collection was selected, move it to that collection
                new_collection_name = None
            else:
                # Make a new collection, use the palette name
                new_collection_name = palette.name
                collection_to_index = None

            palette_to_ramp(
                palette,
                color_ramps=context.scene.color_box_ramps,
                collection_to_index=collection_to_index,
                row_to_index=None,
                new_collection_name=new_collection_name,
            )

        else:
            palettes = collection.palettes
            new_collection_name = collection.name

            # add the first one, which creates a new collection
            palette_to_ramp(
                palettes[0],
                color_ramps=context.scene.color_box_ramps,
                collection_to_index=None,
                row_to_index=None,
                new_collection_name=new_collection_name,
            )

            for palette in palettes[1:]:
                palette_to_ramp(
                    palette,
                    color_ramps=context.scene.color_box_ramps,
                    collection_to_index=len(context.scene.color_box_ramps) - 1,
                    row_to_index=None,
                    new_collection_name=new_collection_name,
                )

        context.area.tag_redraw()
        self.report({"INFO"}, f"Added a new Color Ramp")
        return {"FINISHED"}


class CombineAverageMixin:
    bl_options = {"REGISTER"}
    average_method = "get_linear_rgb_average"

    def get_linear_rgb_average(self, palette, color_count):
        return [
            sum(c) / color_count
            for c in zip(*[c.color for c in palette.colors])
        ]

    def get_linear_hsv_average(self, palette, color_count):
        """
        Ignore the alpha channel,
        take the HSV values from the rgb colors.. calculate a linear average of those HSV values, then convert back
        to rgb (and add 1.0 for the alpha)
        """
        hsv_colors = [
            rgb_to_hsv(*c.color[:3]) for c in palette.colors
        ]  # ignore alpha channel, convert rgb to hsv
        hsv_average = [
            sum(c) / color_count for c in zip(*hsv_colors)
        ]  # average of hsv
        return list(hsv_to_rgb(*hsv_average)) + [1.0]

    def generate_average(self, palette):
        color_count = len(palette.colors)
        new_color = palette.colors.add()
        new_color.color = getattr(self, self.average_method)(
            palette, color_count
        )

        # The average color was added last, so we can remove the first item <color_count> times, and then
        # only the average color is left
        for i in range(color_count):
            palette.colors.remove(0)

    def execute(self, context):
        _type, col_index, row_index = context.scene.color_box_selected_menu

        if _type != 0 or col_index == -1:
            return {"FINISHED"}

        if row_index >= 0:
            palettes = [context.scene.color_box[col_index].palettes[row_index]]
        elif row_index == -1:
            palettes = context.scene.color_box[col_index].palettes

        for palette in palettes:
            self.generate_average(palette)

        context.area.tag_redraw()
        self.report({"INFO"}, "Generated one average color")
        return {"FINISHED"}


class COLORBOX_OT_average_color_rgb(CombineAverageMixin, bpy.types.Operator):
    bl_idname = "color_box.average_color_rgb"
    bl_label = "Average colors"
    bl_description = "Replace the entire row, by one color that is the linear average of rgb values"
    average_method = "get_linear_rgb_average"


class COLORBOX_OT_average_color_hsv(CombineAverageMixin, bpy.types.Operator):
    bl_idname = "color_box.average_color_hsv"
    bl_label = "Mix colors"
    bl_description = "Replace the entire row, by one color that is the average of hsv values"
    average_method = "get_linear_hsv_average"


def clean_string(s):
    """
    Take any string, return a valid Python variable name. Enforce:
    - string has to start with a letter or underscore
    - all other characters have to be letters, numbers, or underscore

    """
    # Remove invalid characters
    s = re.sub("[^0-9a-zA-Z_]", "", s)

    # Remove leading characters until we find a letter or underscore
    s = re.sub("^[^a-zA-Z_]+", "", s)

    return s


class COLORBOX_OT_print_colors(bpy.types.Operator):
    bl_idname = "color_box.print_colors_to_console"
    bl_label = "Print the colors in the Python console"
    bl_description = "Print the colors in the Python console"
    bl_options = {"REGISTER"}

    def get_color_value(
        self,
        weighted_color,
        color_print_type,
        include_aplha=True,
        use_range_255=False,
    ):
        color = weighted_color.color

        if color_print_type == "hsv":
            hsv = list(rgb_to_hsv(*color[:3]))
            if include_aplha:
                hsv += [color[3]]

            if use_range_255:
                return [int(c * 255) for c in hsv]
            else:
                return round_color_tuple(hsv)
        elif color_print_type == "hex":
            return f"'{get_hex_code(color)}'"

        # rgb values as the default
        color_part = color if include_aplha else color[:3]
        if use_range_255:
            return [int(c * 255) for c in color_part]
        else:
            return round_color_tuple(color_part)

    def execute(self, context):
        _type, col_index, row_index = context.scene.color_box_selected_menu

        if _type != 0:
            return {"FINISHED"}

        prefs = context.preferences.addons[PACKAGE_NAME].preferences
        color_print_type = prefs.color_print_type
        include_alpha = prefs.color_print_include_alpha
        use_range_255 = prefs.color_print_use_255_range

        print()

        collection = context.scene.color_box[col_index]
        if row_index == -1:
            print(f"{clean_string(collection.name)}_{color_print_type} = [")
            for palette in collection.palettes:
                print("    [")
                for weighted_color in palette.colors:
                    print(
                        f"        {self.get_color_value(weighted_color, color_print_type, include_alpha, use_range_255)},"
                    )
                print("    ],")
            print("]")
        else:
            palette = collection.palettes[row_index]

            print(f"{clean_string(palette.name)}_{color_print_type} = [")
            for weighted_color in palette.colors:
                print(
                    f"    {self.get_color_value(weighted_color, color_print_type, include_alpha, use_range_255)},"
                )
            print("]")

        print()

        return {"FINISHED"}


class COLORBOX_OT_palette_to_swapper(bpy.types.Operator):
    bl_idname = "colorbox.palette_to_swapper"
    bl_label = "Add to Swapper"
    bl_description = "Add this palette to the Color Swapper"

    def execute(self, context):
        (
            _type,
            collection_index,
            palette_index,
        ) = context.scene.color_box_selected_menu

        if not (collection_index > -1 and palette_index > -1):
            return {"CANCELLED"}

        collection = context.scene.color_box[collection_index]
        palette = collection.palettes[palette_index]
        swapper = context.scene.color_box_swapper

        from .color_swapper import apply_palette

        apply_palette(swapper, palette)

        # by not updating the swapper collection, the previous swapper palette will stay untouched
        # and can be reverted by clicking the button there
        # swapper.selected_color_collection = collection_index
        # swapper.selected_palette = palette_index

        return {"FINISHED"}
