import bpy
import textwrap


class ColorBoxPreferences(bpy.types.AddonPreferences):
    bl_idname = __package__

    active_tab: bpy.props.EnumProperty(
        name="Active tab",
        description="Preferences for this category",
        items=[
            ("colorbox", "Colorbox", ""),
            # ('color_ramp', 'Color Ramp', ''),
            ("color_picker", "Color Picker", ""),
            ("save_load", "Save & load", ""),
            ("info", "Info", ""),
        ],
        default="colorbox",
    )

    #
    # Colorbox tab
    #
    colors_per_row_name = "Colors per row"
    colors_per_row: bpy.props.IntProperty(
        name=colors_per_row_name, default=6, min=2, max=20
    )

    color_row_height_name = "Color row height"
    color_row_height: bpy.props.FloatProperty(
        name=color_row_height_name, default=0.95, min=0.2, max=4.0
    )

    use_weighted_colors_name = "Use weighted Colors"
    use_weighted_colors: bpy.props.BoolProperty(
        name=use_weighted_colors_name,
        default=False,
        description="Weighted Colors for extracted colors",
    )
    analyse_all_pixels_name = "Analyse all pixels"
    analyse_all_pixels: bpy.props.BoolProperty(
        name=analyse_all_pixels_name,
        default=False,
        description="Unchecked means: take a sample of ~2000 pixels only)",
    )
    max_colors_to_return_name = "Maximum number of colors"
    max_colors_to_return: bpy.props.IntProperty(
        name=max_colors_to_return_name,
        default=12,
        min=1,
        max=500,
        description="Maximum number of colors that will be extracted",
    )
    ignore_alpha_below_name = "Ignore transparent colors"
    ignore_alpha_below: bpy.props.FloatProperty(
        name=ignore_alpha_below_name,
        default=1.00,
        min=0.001,
        max=1.0,
        description="Ignore colors with an alpha value below this value",
    )
    pixel_threshold_name = "Use frequency threshold"
    pixel_threshold: bpy.props.IntProperty(
        name=pixel_threshold_name,
        default=800,
        min=0,
        max=2000,
        description="Ignore color values that are this times less frequent then the most frequent pixel",
    )
    determine_distinct_colors_name = "Distinct color value"
    determine_distinct_colors: bpy.props.FloatProperty(
        name=determine_distinct_colors_name,
        default=0.08,
        min=0.0,
        max=1.0,
        description="When this value is non zero, ignore colors within this value of a frequent one",
    )

    gamma_value_name = "Gamma value"
    gamma_value: bpy.props.FloatProperty(
        name=gamma_value_name,
        default=2.4,
        min=0.1,
        max=5.0,
        description="Used in the menu when choosing (inverse) Gamma Correction",
    )

    export_png_folder_name = ".png export output directory"
    export_png_folder: bpy.props.StringProperty(
        name=export_png_folder_name,
        default="",
        description="Extract colors",
        subtype="DIR_PATH",
    )

    color_print_type_name = "Print to console color representation"
    color_print_type: bpy.props.EnumProperty(
        name=color_print_type_name,
        description="Choose which version of the color to print in the Python console",
        items=[
            ("rgb", "RGB", ""),
            ("hsv", "HSV", ""),
            ("hex", "Hexcode", ""),
        ],
        default="rgb",
    )

    color_print_use_255_range_name = "Use range 0 - 255"
    color_print_use_255_range: bpy.props.BoolProperty(
        name=color_print_use_255_range_name, default=False
    )

    color_print_include_alpha_name = "Include alpha value"
    color_print_include_alpha: bpy.props.BoolProperty(
        name=color_print_include_alpha_name, default=False
    )

    #
    # Color Picker tab
    #
    color_picker_scale_name = "Color Picker Scale"
    color_picker_scale: bpy.props.FloatProperty(
        name=color_picker_scale_name, default=1.0, min=0.2, max=5.0
    )

    display_live_hexcodes_name = "Display hexcodes, when moving the mouse"
    display_live_hexcodes: bpy.props.BoolProperty(
        name=display_live_hexcodes_name, default=True
    )

    add_hexcodes_to_colorbox_name = (
        "Display hexcodes on colors after they have been picked"
    )
    add_hexcodes_to_colorbox: bpy.props.BoolProperty(
        name=add_hexcodes_to_colorbox_name, default=False
    )

    #
    # Properties for expanding sections within a tab
    #
    expand_color_box_general: bpy.props.BoolProperty(
        name="Expand the General Colorbox options", default=False
    )
    expand_import_export_options: bpy.props.BoolProperty(
        name="Expand the Import/Export options", default=False
    )
    expand_custom_import_options: bpy.props.BoolProperty(
        name="Expand the Custom Import options", default=False
    )
    expand_color_picker_general: bpy.props.BoolProperty(
        name="Expand options for the Color Picker", default=False
    )
    expand_color_picker_hexcodes: bpy.props.BoolProperty(
        name="Expand Hexcode options for the Color Picker", default=False
    )

    def draw(self, context):
        layout = self.layout
        row = layout.row()

        # row.prop(bpy.context.preferences.addons['colorbox'].preferences, 'active_tab', expand=True)
        row.prop(self, "active_tab", expand=True)
        row = layout.row()
        getattr(self, f"draw_{self.active_tab}")(context, row.box())

    def header_row(self, row, prop, label, emboss=False):
        row.alignment = "LEFT"
        row.prop(
            self,
            prop,
            text="",
            emboss=emboss,
            icon="DISCLOSURE_TRI_RIGHT"
            if not getattr(self, prop)
            else "DISCLOSURE_TRI_DOWN",
        )
        row.prop(self, prop, text=f"{label}", emboss=emboss)

    def draw_prop_box(self, layout, prop_list, top_label=None):
        """
        top_label is an optional text that will be displayed before the box with properties
        """
        if top_label:
            layout.label(text=top_label)

        box = layout.box()
        box.separator()
        for prop_name in prop_list:
            row = box.row()
            row.label(text=getattr(self, f"{prop_name}_name"))
            row.prop(self, prop_name, text="")
        box.separator()

    def draw_section_box(self, layout, expand_prop, section_label, prop_list):
        box = layout.box()
        self.header_row(box.row(align=True), expand_prop, label=section_label)
        if getattr(self, expand_prop):
            self.draw_prop_box(layout=box, prop_list=prop_list)

    def draw_colorbox(self, context, layout):
        layout.label(text="")

        self.draw_section_box(
            layout,
            "expand_color_box_general",
            section_label="Main settings",
            prop_list=["colors_per_row", "color_row_height", "gamma_value"],
        )
        self.draw_section_box(
            layout,
            "expand_import_export_options",
            section_label="Import / Export options",
            prop_list=[
                # Export .png options
                "export_png_folder",
                # 'export_png_width', 'export_png_height',
                # Print to console options
                "color_print_type",
                "color_print_use_255_range",
                "color_print_include_alpha",
            ],
        )
        self.draw_section_box(
            layout,
            "expand_custom_import_options",
            section_label="Custom Import",
            prop_list=[
                "analyse_all_pixels",
                "max_colors_to_return",
                "ignore_alpha_below",
                "pixel_threshold",
                "determine_distinct_colors",
                "use_weighted_colors",
            ],
        )

    def draw_color_ramp(self, context, layout):
        layout.label(text="This is the Color Ramp tab")

    def draw_color_picker(self, context, layout):
        layout.label(
            text="The Color Picker can be launched in the Image Editor with Ctrl + Shift + P"
        )
        self.draw_section_box(
            layout,
            "expand_color_picker_general",
            section_label="General settings",
            prop_list=["color_picker_scale"],
        )
        self.draw_section_box(
            layout,
            "expand_color_picker_hexcodes",
            section_label="Hexcodes",
            prop_list=["display_live_hexcodes", "add_hexcodes_to_colorbox"],
        )

    def draw_info(self, context, layout):
        col = layout.column(align=True)
        text = """
        本插件由XZ-Blender修复并汉化!
        感谢您的下载与支持!
        """
        ui_scale = context.preferences.view.ui_scale
        width = 90 / ui_scale
        for paragraph in text.split("\n"):
            for line in textwrap.wrap(paragraph, width):
                col.label(text=line.lstrip())

        row = layout.row()
        col1 = row.column()
        col2 = row.column()
        # col1.operator(
        #     "wm.url_open", text="Essential Elephant", icon="URL"
        # ).url = "https://www.essentialelephant.com/"
        # col2.operator(
        #     "wm.url_open", text="Support Blender", icon="FUND"
        # ).url = "https://www.blender.org/foundation/donation-payment/"

    def draw_save_load(self, context, layout):
        layout.operator(
            "color_box.extract_from_image", text="Load Colorbox from image"
        )

        if context.scene.color_box:
            # Only display the save Colorbox button if there is something to save
            layout.operator(
                "color_box.save_json", text="Save Colorbox to disk"
            )

        layout.operator(
            "color_box.import_blender_palettes", text="Import Blender Palettes"
        )
