import bpy
from .utils.move_rows import MoveRow
from . import utils


MOVE_DESCRIPTION = "After clicking on this button, click on the button where it should move to"


def register():
    bpy.types.Scene.color_box_mode = bpy.props.EnumProperty(
        name="Color Mode",
        description="Select the type of color formation",
        items=[
            ("Triad", "Triad", ""),
            ("Tetradic", "Tetradic", ""),
            ("Analogous", "Analogous", ""),
            ("Complementary", "Complementary", ""),
            ("Split-Complementary", "Split-Complementary", ""),
            ("Monochromatic", "Monochromatic", ""),
        ],
        default="Complementary",
        update=update_colors,
    )
    
    bpy.types.Scene.color_box_scheme_color_1 = bpy.props.FloatVectorProperty(
        subtype="COLOR_GAMMA",
        min=0.0,
        max=1.0,
        size=4,
        default=[1, 0, 0, 1],
        update=update_colors,
    )
    bpy.types.Scene.color_box_scheme_color_2 = bpy.props.FloatVectorProperty(
        subtype="COLOR_GAMMA",
        min=0.0,
        max=1.0,
        size=4,
        default=[0, 1, 1, 1],
    )
    bpy.types.Scene.color_box_scheme_color_3 = bpy.props.FloatVectorProperty(
        subtype="COLOR_GAMMA",
        min=0.0,
        max=1.0,
        size=4,
        default=[0.5, 0, 1, 0],
    )
    bpy.types.Scene.color_box_scheme_color_4 = bpy.props.FloatVectorProperty(
        subtype="COLOR_GAMMA",
        min=0.0,
        max=1.0,
        size=4,
        default=[0.5, 1, 0, 0],
    )


def update_colors(self, context):
    # print("color 1 updated:", self, context)
    color_amount = 3
    if context.scene.color_box_mode == "Monochromatic":
        color_amount = 1
    elif context.scene.color_box_mode == "Complementary":
        color_amount = 2
    elif context.scene.color_box_mode == "Tetradic":
        color_amount = 4
    
    color_1 = context.scene.color_box_scheme_color_1
    colors_fields = [context.scene.color_box_scheme_color_2, context.scene.color_box_scheme_color_3, context.scene.color_box_scheme_color_4]
    
    # Set the alpha of each color to 1 or 0 depending if they are used in the mode
    for i, color in enumerate(colors_fields):
        if i < color_amount - 1:
            color[3] = 1
        else:
            color[3] = 0
    
    # Set the colors
    colors = []
    if context.scene.color_box_mode == "Monochromatic":
        return
    elif context.scene.color_box_mode == "Complementary":
        colors = [utils.blender.complementary(color_1)]
    elif context.scene.color_box_mode == "Split-Complementary":
        colors = utils.blender.split_complementary(color_1)
    elif context.scene.color_box_mode == "Triad":
        colors = utils.blender.triad(color_1)
    elif context.scene.color_box_mode == "Tetradic":
        colors = utils.blender.tetradic(color_1)
    elif context.scene.color_box_mode == "Analogous":
        colors = utils.blender.analogous(color_1)

    # Apply the colors
    for i, color in enumerate(colors):
        colors_fields[i][0] = color[0]
        colors_fields[i][1] = color[1]
        colors_fields[i][2] = color[2]


def move_individual_color(self, context):
    from_indices = -1, -1, -1
    to_indices = -1, -1, -1
    color_box = context.scene.color_box

    for collection_index, collection in enumerate(color_box):
        for row_index, palette in enumerate(collection.palettes):
            for color_index, weighted_color in enumerate(palette.colors):
                if weighted_color.selected:
                    if weighted_color != self:
                        from_indices = collection_index, row_index, color_index
                    else:
                        to_indices = collection_index, row_index, color_index

    # if any single one of the indices equals -1, do nothing
    if any([_index == -1 for _index in from_indices + to_indices]):
        return

    from_index_color = from_indices[2]
    to_palette = color_box[to_indices[0]].palettes[to_indices[1]]
    from_palette = color_box[from_indices[0]].palettes[from_indices[1]]
    from_palette.colors[from_indices[2]].selected = False
    to_palette.colors[to_indices[2]].selected = False

    if not (
        from_indices[0] == to_indices[0] and from_indices[1] == to_indices[1]
    ):
        # the 2 colors are not in the same row. Add a new one to the to row, remove the color from the other row
        from_color = from_palette.colors[from_indices[2]]

        new_color = to_palette.colors.add()
        new_color.weight = from_color.weight
        new_color.color = from_color.color

        from_palette.colors.remove(from_indices[2])

        from_index_color = (
            len(color_box[to_indices[0]].palettes[to_indices[1]].colors) - 1
        )

    # move a color in the same row
    color_box[to_indices[0]].palettes[to_indices[1]].colors.move(
        from_index_color, to_indices[2]
    )

    context.area.tag_redraw()


class WeightedColor(bpy.types.PropertyGroup):
    color: bpy.props.FloatVectorProperty(
        subtype="COLOR_GAMMA", name="", min=0.0, max=1.0, size=4
    )
    weight: bpy.props.FloatProperty(default=0.09)
    selected: bpy.props.BoolProperty(
        default=False, update=move_individual_color
    )


def move_palette(self, context):
    MoveRow(thing=self, context=context, collection_type_to=0)


def move_ramp(self, context):
    MoveRow(thing=self, context=context, collection_type_to=1)


class Palette(bpy.types.PropertyGroup):
    name: bpy.props.StringProperty(name="palette name")
    colors: bpy.props.CollectionProperty(type=WeightedColor)
    selected: bpy.props.BoolProperty(
        default=False,
        update=move_palette,
        name="Move",
        description=MOVE_DESCRIPTION,
    )
    show_color_selection: bpy.props.BoolProperty(
        default=False, name="Show color ordering buttons"
    )


class PaletteCollection(bpy.types.PropertyGroup):
    name: bpy.props.StringProperty(name="collection name")
    palettes: bpy.props.CollectionProperty(type=Palette)
    selected_row: bpy.props.IntProperty(default=-1)
    show: bpy.props.BoolProperty(default=False)
    selected: bpy.props.BoolProperty(
        default=False,
        update=move_palette,
        name="Move",
        description=MOVE_DESCRIPTION,
    )
    show_color_selection: bpy.props.BoolProperty(
        default=False, name="Show color ordering buttons for all rows"
    )
    show_palette_options: bpy.props.BoolProperty(
        default=False, name="Show / hide options for every palette"
    )


class ColorRamp(bpy.types.PropertyGroup):
    name: bpy.props.StringProperty(name="color ramp name")
    selected: bpy.props.BoolProperty(
        default=False,
        update=move_ramp,
        name="Move",
        description=MOVE_DESCRIPTION,
    )
    # json_data: bpy.props.StringProperty()


class ColorRampCollection(bpy.types.PropertyGroup):
    name: bpy.props.StringProperty(name="collection name")
    ramps: bpy.props.CollectionProperty(type=ColorRamp)
    show: bpy.props.BoolProperty(default=False)
    selected: bpy.props.BoolProperty(
        default=False,
        update=move_ramp,
        name="Move",
        description=MOVE_DESCRIPTION,
    )


def color_ramp_enum_callback(scene, context):
    items = [
        ("-2", "Select a collection below:", ""),
        ("-1", "Add to a new Collection", ""),
    ]

    for index, color_ramp_collection in enumerate(
        bpy.context.scene.color_box_ramps
    ):
        items.append(
            (str(index), f"Add to '{color_ramp_collection.name}'", "")
        )

    return items


def palettes_enum_callback(scene, context):
    items = [
        ("-2", "Select a collection below:", ""),
        ("-1", "Add to a new Collection", ""),
    ]

    for index, collection in enumerate(bpy.context.scene.color_box):
        items.append((str(index), f"Add to '{collection.name}'", ""))

    return items


def effectors_callback(scene, context):
    return [
        ("reverse_colors", "Reverse palette order", ""),
        ("apply_gamma", "Gamma correction", ""),
        ("apply_inverse_gamma", "Inverse gamma correction", ""),
        ("sort_by_hue", "Sort by hue", ""),
        ("average_color_rgb", "Average colors", ""),
        ("average_color_hsv", "Mix colors", ""),
        ("print_colors_to_console", "Print to console", ""),
    ]


def update_effectors(self, context):
    getattr(bpy.ops.color_box, self.selected_effector)()


def save_color_ramp_as_palette(self, context):
    if self.selected_palette != "-2":
        bpy.ops.color_box.color_ramp_convert_to_color_row()


def save_palette_as_color_ramp(self, context):
    if self.selected_color_ramp != "-2":
        bpy.ops.color_box.add_to_color_ramp()


class ColorBoxOptions(bpy.types.PropertyGroup):
    selected_color_ramp: bpy.props.EnumProperty(
        name="Color Ramp Collection",
        description="Select the Color Ramp Collection",
        items=color_ramp_enum_callback,
        update=save_palette_as_color_ramp,
    )
    selected_palette: bpy.props.EnumProperty(
        name="Palette Collection",
        description="Select the Palette Collection",
        items=palettes_enum_callback,
        update=save_color_ramp_as_palette,
    )
    selected_effector: bpy.props.EnumProperty(
        name="Select and apply Effector",
        description="Select one of the effectors from the list, and apply it",
        items=effectors_callback,
        update=update_effectors,
    )


def set_connected_colors(self, context):
    new_color = tuple(self.color)

    for item in self.color_pointers:
        socket = item.resolved_socket
        if socket is not None:
            setattr(socket, item.property_name, new_color)


class ColorFieldPointProperty(
    bpy.types.PropertyGroup
):  # one item in the ColorField refs
    material: bpy.props.PointerProperty(
        name="material", type=bpy.types.Material
    )
    world_material: bpy.props.PointerProperty(
        name="material", type=bpy.types.World
    )
    light_material: bpy.props.PointerProperty(
        name="material", type=bpy.types.Light
    )
    node_group_name: bpy.props.StringProperty()
    node_name: bpy.props.StringProperty()
    property_name: bpy.props.StringProperty()
    socket_index: bpy.props.IntProperty()
    pointer_id: bpy.props.StringProperty()
    inputs_outputs: bpy.props.StringProperty(default="inputs")

    @property
    def resolved_node(self):
        if self.node_group_name:
            return bpy.data.node_groups[self.node_group_name].nodes[
                self.node_name
            ]

        if self.material:
            material = self.material
        elif self.world_material:
            material = self.world_material
        elif self.light_material:
            material = self.light_material

        return material.node_tree.nodes[self.node_name]

    @property
    def resolved_socket(self):
        node = self.resolved_node
        if self.inputs_outputs == "elements":
            node = getattr(node, "color_ramp")

        return getattr(node, self.inputs_outputs)[self.socket_index]

    @property
    def resolved_socket_name(self):
        if self.inputs_outputs == "elements":
            return f"color{self.socket_index}"

        return (
            self.resolved_socket.name
            if self.resolved_socket is not None
            else None
        )

    @property
    def resolved_socket_property(self):
        return getattr(self.resolved_socket, self.property_name)

    @property
    def resolved_users_of_material(self):
        objects = []
        for obj in bpy.data.objects:
            for slot in obj.material_slots:
                if slot.material == self.material:
                    objects.append(obj)
        return objects

    @property
    def resolved_objects_name(self):
        if self.world_material:
            return "World"

        return ", ".join([obj.name for obj in self.resolved_users_of_material])


class ColorPointersCollection(bpy.types.PropertyGroup):
    active_index: bpy.props.IntProperty()
    color_pointers: bpy.props.CollectionProperty(
        name="Object Collection", type=ColorFieldPointProperty
    )
    color: bpy.props.FloatVectorProperty(
        subtype="COLOR",
        name="",
        min=0.0,
        max=1.0,
        size=4,
        default=(1.0, 1.0, 1.0, 1.0),
        update=set_connected_colors,
    )


class ColorSwapCollection(bpy.types.PropertyGroup):
    active_index: bpy.props.IntProperty()
    color_collection: bpy.props.CollectionProperty(
        name="Name Collection", type=ColorPointersCollection
    )
    selected_color_collection: bpy.props.IntProperty(default=0)
    selected_palette: bpy.props.IntProperty(default=0)
    auto_apply: bpy.props.BoolProperty(
        default=False, 
        description="Option to apply the palette directly when navigating"
    )
    auto_add_swaps: bpy.props.BoolProperty(
        default=False, 
        description="Option to add additional color swap slots directly when navigating"
    )
