import bpy

from .utils.color_ramp import ColorBoxMaterial
from . import operators


GRIP_ICON = "GRIP"  # 'HANDLETYPE_VECTOR_VEC'
CHECKBOX_ICON = "CLIPUV_DEHLT"


class ColorBoxMixin:
    bl_region_type = "UI"
    bl_category = "Colorbox"
    bl_label = "PALETTES"

    collection_type = 0  # 0 for palette, 1 for color_ramp

    def draw(self, context):
        scene = context.scene

        addon_prefs = context.preferences.addons[__package__].preferences
        colors_per_row = addon_prefs.colors_per_row
        color_row_height = addon_prefs.color_row_height
        use_weighted_colors = addon_prefs.use_weighted_colors

        row = self.layout.row()  # align=True)
        row.scale_y = 1.7
        col1 = row.column()
        col2 = row.column()
        col3 = row.column()

        # from . import custom_icons
        # global custom_icons

        col1.operator(
            "color_box.collection_add", text="", icon="PLUS"
        ).collection_type = self.collection_type
        col2.operator("color_box.extract_from_palette", text="", icon="COLOR")
        col3.operator("color_box.load_json", text="", icon="IMPORT")

        col1.scale_x = 100.0
        col2.scale_x = 100.0
        col3.scale_x = 100.0

        # selected_row = context.scene.color_box_selected_row
        for collection_index, collection in enumerate(context.scene.color_box):
            collection_row = self.layout.row(align=True)

            show_col_op = collection_row.operator(
                "color_box.collection_show",
                text="",  # emboss=False
                icon="DOWNARROW_HLT" if collection.show else "RIGHTARROW",
            )

            collection_row.prop(
                collection, "name", text=""
            )  # display the palette name in an editable field
            remove_op = collection_row.operator(
                "color_box.collection_remove", text="", icon="CANCEL"
            )
            collection_row.prop(
                collection,
                "show_palette_options",
                text="",
                icon="RADIOBUT_ON"
                if collection.show_palette_options
                else "RADIOBUT_OFF",
            )
            col_menu_op = collection_row.operator(
                "color_box.show_menu", icon="PREFERENCES", text=""
            )

            for _op in [show_col_op, remove_op, col_menu_op]:
                _op.collection_type = self.collection_type
                _op.collection_index = collection_index
                _op.row_index = -1

            collection_row.prop(
                collection, "selected", text="", icon=GRIP_ICON
            )

            if not collection.show:
                continue

            current_page = scene.color_box_palette_current_page
            palette_per_page = scene.color_box_palette_per_page

            page_from = palette_per_page * current_page
            page_to = page_from + palette_per_page

            for row_index, palette in enumerate(
                collection.palettes[page_from:page_to]
            ):
                box = self.layout.box()

                if collection.show_palette_options:
                    label_row = box.row(align=True)

                    label_row.prop(
                        palette, "name", text=""
                    )  # display the palette name in an editable field
                    # label_row.label(text=palette.get('name'))  # alternative: label with an edit icon / toggle icon

                    remove_row_op = label_row.operator(
                        "color_box.remove_color_row", text="", icon="CANCEL"
                    )
                    remove_color_op = label_row.operator(
                        "color_box.remove_color", icon="REMOVE", text=""
                    )
                    add_color_op = label_row.operator(
                        "color_box.add_color", icon="ADD", text=""
                    )
                    show_menu_op = label_row.operator(
                        "color_box.show_menu", icon="PREFERENCES", text=""
                    )

                    for ui_operator in [
                        remove_row_op,
                        remove_color_op,
                        add_color_op,
                        show_menu_op,
                    ]:
                        ui_operator.row_index = row_index
                        ui_operator.collection_index = collection_index

                    show_menu_op.collection_type = self.collection_type

                    label_row.prop(
                        palette, "selected", text="", icon=GRIP_ICON
                    )
                    label_row.scale_y = 0.9

                color_row = box.row(align=True)

                # color_row.enabled = False  # disabling makes the color not dragable
                for _color_index, weighted_color in enumerate(
                    palette.get("colors", [])
                ):
                    # add a new row after every 6 items. Only do it if there is at least one more color to come
                    curr_color = palette.colors[_color_index]
                    if (
                        _color_index > (colors_per_row - 1)
                        and _color_index % colors_per_row == 0
                    ):
                        color_row = box.row(align=True)

                    row = color_row.row(align=True)
                    col = row.column(align=True)

                    if (
                        palette.show_color_selection
                        or collection.show_color_selection
                    ):
                        col.prop(
                            curr_color, "selected", text="", icon=GRIP_ICON
                        )
                    col.prop(curr_color, "color")

                    if use_weighted_colors:
                        row.scale_x = curr_color
                    row.scale_y = color_row_height

            row = self.layout.row(align=True)
            row.alignment = "CENTER"

            row.operator(
                "color_box.previous_palette_page", text="", icon="TRIA_LEFT"
            )
            row.operator(
                "color_box.next_palette_page", text="", icon="TRIA_RIGHT"
            )
            row = self.layout.row(align=True)
            row.scale_y = 1.7
            row.operator(
                "color_box.add_color_row", text="Add row"
            ).collection_index = collection_index
            self.layout.separator()


class ColorRampMixin:
    bl_region_type = "UI"
    bl_category = "Colorbox"
    bl_label = "COLOR RAMPS"
    bl_options = {"DEFAULT_CLOSED"}

    collection_type = 1  # 0 for palette, 1 for color_ramp

    def draw(self, context):
        for collection_index, collection in enumerate(
            context.scene.color_box_ramps
        ):
            collection_row = self.layout.row()

            show_col_op = collection_row.operator(
                "color_box.collection_show",
                text="",  # emboss=False
                icon="DOWNARROW_HLT" if collection.show else "RIGHTARROW",
            )
            collection_row.prop(
                collection, "name", text=""
            )  # display the color_ramp name in an editable field
            remove_op = collection_row.operator(
                "color_box.collection_remove", text="", icon="CANCEL"
            )
            col_menu_op = collection_row.operator(
                "color_box.show_menu", icon="PREFERENCES", text=""
            )

            for _op in [show_col_op, remove_op, col_menu_op]:
                _op.collection_type = self.collection_type
                _op.collection_index = collection_index
                _op.row_index = -1

            collection_row.prop(
                collection, "selected", text="", icon=GRIP_ICON
            )

            if not collection.show:
                continue

            for row_index, color_ramp in enumerate(collection.ramps):
                try:
                    color_ramp_node = ColorBoxMaterial.get_node_by_name(
                        color_ramp.name
                    )
                except KeyError:
                    # f'Color Ramp node not found')  # Make sure this can never happen. Names have to be in sync
                    # better: names should not matter... scene.property should point to an node identifier (not the name
                    continue

                box = self.layout.box()

                label_row = box.row(align=True)

                # BACK   TRIA_LEFT   ADD   (alternative Icons)
                label_row.operator(
                    "color_box.color_ramp_load", icon="TRIA_LEFT", text=""
                ).row_index = row_index
                # this should work with a update or get/set structure, where changing the name, will change the ghost
                # node name, and the actual name from the node will be returned (use can add 'ColorRamp' and the name
                # will be set to 'ColorRamp.003') pointer property seemed nice, so the text label here is the one from
                # the node. But that seems impossible unfortunately. Alternative: make a custom name field that shadows
                # the actual node name
                # label_row.prop(color_ramp, 'name', text='')  # this does not work, changing the name messes things up
                label_row.label(text=color_ramp.name)
                remove_op = label_row.operator(
                    "color_box.color_ramp_remove", text="", icon="CANCEL"
                )
                remove_op.row_index = row_index
                remove_op.collection_index = collection_index

                show_menu_op = label_row.operator(
                    "color_box.show_menu", icon="PREFERENCES", text=""
                )
                show_menu_op.row_index = row_index
                show_menu_op.collection_index = collection_index
                show_menu_op.collection_type = self.collection_type

                label_row.prop(color_ramp, "selected", text="", icon=GRIP_ICON)

                row = box.box()  # align=True)
                row.template_color_ramp(
                    color_ramp_node, "color_ramp", expand=True
                )

            row = self.layout.row(align=True)
            row.scale_y = 1.7
            row.operator(
                "color_box.color_ramp_add", text="Add Color Ramp"
            ).collection_index = collection_index
            self.layout.separator()

        row = self.layout.row()
        row.scale_y = 1.7
        col1 = row.column()
        col1.scale_x = 100.0
        col1.operator(
            "color_box.collection_add", text="", icon="PLUS"
        ).collection_type = self.collection_type
        

class ColorSchemeMixin:
    bl_region_type = "UI"
    bl_category = "Colorbox"
    bl_label = "COLOR SCHEMES"
    bl_options = {"DEFAULT_CLOSED"}

    def draw(self, context):
        layout = self.layout

        row = layout.row()
        row.prop(context.scene, "color_box_mode", text="Mode")
        row.operator(operators.color_scheme.COLORBOX_OT_color_scheme_help.bl_idname, text="", icon="QUESTION")

        row = layout.row()
        row.scale_y = 1.3
        col = row.column()
        col.prop(context.scene, "color_box_scheme_color_1", text="")

        # Get amount of colors to display, depending on their alpha value
        color_amount = 0
        color_ids = ["color_box_scheme_color_2", "color_box_scheme_color_3", "color_box_scheme_color_4"]
        for color_id in color_ids:
            color = getattr(context.scene, color_id)
            if color[3] > 0:
                color_amount += 1

        if color_amount > 0:
            col = row.column()
            col.scale_x = 0.25
            col.label(text=">")

            for color_id in color_ids:
                color = getattr(context.scene, color_id)
                if color[3] > 0:
                    col = row.column()
                    col.prop(context.scene, color_id, text="")
        
        row = layout.row()
        row.scale_y = 1.7
        row.operator(operators.color_scheme.COLORBOX_OT_color_scheme_add_to_collection.bl_idname, text="Add to Collection")
        
        
class COLORBOX_UL_swap_collection(bpy.types.UIList):
    def draw_item(
        self,
        context,
        layout,
        data,
        item,
        icon,
        active_data,
        active_propname,
        index,
    ):
        split = layout.split(factor=0.25)
        split.prop(item, "color", text="")
        # split = split.split(factor=0.15)  # attemt to get rid of the colon  'index:', but the split in 2 is nicer
        # split.label(text=str(index))
        split.prop(
            item, "name", text=str(index), emboss=False, translate=False
        )  # icon="USER")

    def invoke(self, context, event):
        pass


class COLORBOX_UL_swap_color(bpy.types.UIList):
    def draw_item(
        self,
        context,
        layout,
        data,
        item,
        icon,
        active_data,
        active_propname,
        index,
    ):
        split = layout.split(factor=0.1)
        split.prop(data, "color", text="")
        split.enabled = False

        split.label(text=f"{item.resolved_objects_name}")
        split.label(text=f"{item.node_name}")
        socket_name = item.resolved_socket_name
        if socket_name is None:
            split.label(text="", icon="ERROR")
        else:
            split.label(text=socket_name)

    def invoke(self, context, event):
        pass


class SwapPanelMixin:
    bl_region_type = "UI"
    bl_category = "Colorbox"
    bl_label = "COLOR SWAPPER"
    bl_space_type = "VIEW_3D"
    bl_options = {"DEFAULT_CLOSED"}

    def draw_palette_navigattion(self, layout, swap_collection, color_box):
        col_index = swap_collection.selected_color_collection

        if not (0 <= col_index < len(color_box)):
            return

        color_collection = color_box[col_index]

        # add lbr, center_part and rbr all in collection_row.... Only to achieve a centered label in the center part...
        collection_row = layout.row()
        lbr = collection_row.row()
        left_button = lbr.operator(
            "colorbox.swap_colorbox_navigation", icon="TRIA_LEFT", text=""
        )
        left_button.navigation_type = (0, -1)

        center_part = collection_row.row()
        center_part.alignment = "CENTER"
        center_part.label(text=color_collection.name)

        rbr = collection_row.row()
        right_button = rbr.operator(
            "colorbox.swap_colorbox_navigation", icon="TRIA_RIGHT", text=""
        )
        right_button.navigation_type = (0, 1)

        pal_index = swap_collection.selected_palette
        if not (0 <= pal_index < len(color_collection.palettes)):
            return
        palette = color_collection.palettes[swap_collection.selected_palette]

        palette_name_row = layout.row()
        palette_name_row.alignment = "CENTER"
        palette_name_row.label(text=palette.name)
        palette_name_row.scale_y = 0.4

        palette_row = layout.row(align=True)
        palette_row.scale_y = 1.5
        left_button = palette_row.operator(
            "colorbox.swap_colorbox_navigation", icon="TRIA_LEFT", text=""
        )
        for _color_index, weighted_color in enumerate(palette.colors[:16]):
            palette_row.prop(palette.colors[_color_index], "color")
        right_button = palette_row.operator(
            "colorbox.swap_colorbox_navigation", icon="TRIA_RIGHT", text=""
        )
        left_button.navigation_type = (1, -1)
        right_button.navigation_type = (1, 1)

    def draw_button_row_(self, layout, swap_collection, color_box):
        button_row = layout.row()
        button_row.alignment = "CENTER"
        button_row.scale_x = 2

        if len(color_box):
            col = button_row.column()
            col.operator(
                "colorbox.swap_apply_palette", icon="FILE_REFRESH", text=""
            )
            col = button_row.column()
            col.prop(
                swap_collection,
                "auto_apply",
                text="",
                icon="LINKED" if swap_collection.auto_apply else "UNLINKED",
            )
            col = button_row.column()
            col.enabled = swap_collection.auto_apply
            col.prop(
                swap_collection,
                "auto_add_swaps",
                text="",
                icon="SNAP_ON" if swap_collection.auto_add_swaps else "SNAP_OFF",
            )

        col = button_row.column()
        col.operator("colorbox.swap_shuffle", icon="SHADERFX", text="")

        col = button_row.column()
        col.operator(
            "colorbox.clear_all_swappers", icon="CANCEL", text=""
        )

    def draw(self, context):
        layout = self.layout.box()
        swap_collection = context.scene.color_box_swapper
        color_box = context.scene.color_box

        self.draw_palette_navigattion(layout, swap_collection, color_box)
        self.draw_button_row_(layout, swap_collection, color_box)

        row = layout.row()
        row.template_list(
            listtype_name="COLORBOX_UL_swap_collection",
            list_id="",
            dataptr=swap_collection,
            propname="color_collection",
            active_dataptr=swap_collection,
            active_propname="active_index",
            rows=7,
        )

        col = row.column()
        col.operator("colorbox.swap_collection_add", text="", icon="ADD")
        col.operator("colorbox.swap_collection_delete", text="", icon="REMOVE")

        col.separator()

        col.operator(
            "color_box.create_swapper_nodes", text="", icon="NODE"
        )

        row = layout.row()
        if (
            len(swap_collection.color_collection) > 0
            and swap_collection.active_index != -1
        ):
            color_user = swap_collection.color_collection[
                swap_collection.active_index
            ]

            row.template_list(
                listtype_name="COLORBOX_UL_swap_color",
                list_id="",
                dataptr=color_user,
                propname="color_pointers",
                active_dataptr=color_user,
                active_propname="active_index",
                rows=4,
            )
            col = row.column()
            col.operator(
                "colorbox.swap_color_pointer_delete", text="", icon="REMOVE"
            )


class VIEW3D_PT_swap_panel(SwapPanelMixin, bpy.types.Panel):
    bl_space_type = "VIEW_3D"


class NODE_EDITOR_PT_swap_panel(SwapPanelMixin, bpy.types.Panel):
    bl_space_type = "NODE_EDITOR"


class VIEW3D_PT_color_box(ColorBoxMixin, bpy.types.Panel):
    bl_space_type = "VIEW_3D"


class NODE_EDITOR_PT_color_box(ColorBoxMixin, bpy.types.Panel):
    bl_space_type = "NODE_EDITOR"


class IMAGE_EDITOR_PT_color_box(ColorBoxMixin, bpy.types.Panel):
    bl_space_type = "IMAGE_EDITOR"


class VIEW3D_PT_color_ramps(ColorRampMixin, bpy.types.Panel):
    bl_space_type = "VIEW_3D"


class NODE_EDITOR_PT_color_ramps(ColorRampMixin, bpy.types.Panel):
    bl_space_type = "NODE_EDITOR"


class IMAGE_EDITOR_PT_color_ramps(ColorRampMixin, bpy.types.Panel):
    bl_space_type = "IMAGE_EDITOR"


class VIEW3D_PT_color_schemes(ColorSchemeMixin, bpy.types.Panel):
    bl_space_type = "VIEW_3D"


class NODE_EDITOR_PT_color_schemes(ColorSchemeMixin, bpy.types.Panel):
    bl_space_type = "NODE_EDITOR"


class IMAGE_EDITOR_PT_color_schemes(ColorSchemeMixin, bpy.types.Panel):
    bl_space_type = "IMAGE_EDITOR"
