from colorsys import rgb_to_hsv, hsv_to_rgb


def get_material(context):
    """
    Return the material from the context,
    and if it does not exist, return the first material from the selected object(s).

    Return None if none of the above could be found
    """
    if hasattr(context, "material") and context.material is not None:
        return context.material
    elif hasattr(context, "selected_objects"):
        for obj in context.selected_objects:
            for mat_slot in obj.material_slots:
                if mat_slot.material and mat_slot.material is not None:
                    return mat_slot.material
    # returning None now. We could create a material for a selected object without a material
    # material = bpy.data.materials.new('NewMaterial')


def complementary(color):
    """returns RGB components of complementary color"""
    hsv = rgb_to_hsv(color[0], color[1], color[2])
    return hsv_to_rgb((hsv[0] + 0.5) % 1, hsv[1], hsv[2])


def split_complementary(color):
    """returns RGB components of split-complementary colors"""
    hsv = rgb_to_hsv(color[0], color[1], color[2])
    hue = hsv[0]

    shift = 1 / 12

    hue1 = hue + 0.5 + shift
    hue2 = hue + 0.5 - shift

    color1 = hsv_to_rgb(hue1 % 1, hsv[1], hsv[2])
    color2 = hsv_to_rgb(hue2 % 1, hsv[1], hsv[2])

    return color1, color2


def analogous(color):
    """returns RGB components of analogous colors"""
    hsv = rgb_to_hsv(color[0], color[1], color[2])
    hue = hsv[0]

    shift = 1 / 12

    hue1 = hue + shift
    hue2 = hue - shift

    color1 = hsv_to_rgb(hue1 % 1, hsv[1], hsv[2])
    color2 = hsv_to_rgb(hue2 % 1, hsv[1], hsv[2])

    return color1, color2


def triad(color):
    """returns RGB components of triadic colors"""
    hsv = rgb_to_hsv(color[0], color[1], color[2])
    hue = hsv[0]

    shift = 2 / 12

    hue1 = hue + 0.5 + shift
    hue2 = hue + 0.5 - shift

    color1 = hsv_to_rgb(hue1 % 1, hsv[1], hsv[2])
    color2 = hsv_to_rgb(hue2 % 1, hsv[1], hsv[2])

    return color1, color2


def tetradic(color):
    """returns RGB components of tetradic colors"""
    hsv = rgb_to_hsv(color[0], color[1], color[2])
    hue = hsv[0]

    hue1 = hue + 0.5
    hue2 = hue - 0.25
    hue3 = hue + 0.25

    color1 = hsv_to_rgb(hue1 % 1, hsv[1], hsv[2])
    color2 = hsv_to_rgb(hue2 % 1, hsv[1], hsv[2])
    color3 = hsv_to_rgb(hue3 % 1, hsv[1], hsv[2])

    return color1, color2, color3
