import bpy
import json


def sort_key(vec):
    return vec.x, vec.y


def determine_node_location(material, cursor_location=None):
    if cursor_location:
        return cursor_location[0] - 300, cursor_location[1]

    min_node_location = min(
        [node.location for node in material.node_tree.nodes], key=sort_key
    )
    return min_node_location.x - 200, min_node_location.y


def add_node_from_data(
    ramp_data, material, new_node_location=None, cursor_location=None
):
    material_name = material.name
    material.use_nodes = True

    if new_node_location is None:
        new_node_location = determine_node_location(
            material, cursor_location=cursor_location
        )

    color_ramp_node = material.node_tree.nodes.new("ShaderNodeValToRGB")
    color_ramp_node.location = new_node_location

    build_color_ramp_node(color_ramp_node=color_ramp_node, ramp_data=ramp_data)

    return material_name, color_ramp_node.name


def build_color_ramp_node(color_ramp_node, ramp_data):
    color_ramp_node.name = ramp_data.get("name", "ColorRamp")
    color_ramp_node.color = ramp_data.get("color", (0, 0, 0))
    color_ramp_node.inputs[0].default_value = ramp_data.get("input_fac", 0.5)
    color_ramp_node.use_custom_color = ramp_data.get("use_custom_color", False)

    color_ramp = color_ramp_node.color_ramp
    color_ramp.color_mode = ramp_data.get("color_mode", "RGB")
    color_ramp.interpolation = ramp_data.get("interpolation", "LINEAR")
    color_ramp.hue_interpolation = ramp_data.get("hue_interpolation", "NEAR")

    for index, position, rgba_color_tuple, alpha in ramp_data["elements"]:
        if index < 2:
            elem = color_ramp.elements[index]
        else:
            elem = color_ramp.elements.new(position=index)

        elem.position = position
        elem.color = rgba_color_tuple
        elem.alpha = alpha

        if index == 31:  # color ramp cannot have more then 32 colors
            return


def get_color_ramp_node_dict(color_ramp_node):
    float_precision = 4
    color_ramp = color_ramp_node.color_ramp
    inputs = color_ramp_node.inputs
    input_fac = inputs[0].default_value if inputs and len(inputs) == 1 else 1.0

    return {
        "name": color_ramp_node.name,
        "color": tuple(
            round(cv, float_precision) for cv in color_ramp_node.color
        ),  # len 3
        "input_fac": round(input_fac, float_precision),
        "use_custom_color": color_ramp_node.use_custom_color,
        "color_mode": color_ramp.color_mode,  # 'RGB', 'HSV', 'HSL'
        # enum in [‘EASE’, ‘CARDINAL’, ‘LINEAR’, ‘B_SPLINE’, ‘CONSTANT’], default ‘LINEAR’
        "interpolation": color_ramp.interpolation,
        # enum in [‘NEAR’, ‘FAR’, ‘CW’, ‘CCW’], default ‘NEAR’
        "hue_interpolation": color_ramp.hue_interpolation,
        "elements": [
            (
                index,
                round(element.position, float_precision),
                tuple(
                    round(cv, float_precision) for cv in element.color
                ),  # this one has length 4, including alpha
                round(element.alpha, float_precision),
            )
            for index, element in enumerate(color_ramp.elements)
        ],
    }


class ColorBoxMaterial:
    """
    This is a 'ghost material', one that should not be used by the user for any objects.
    It just hold a collection of ColorRamps, so they can be used in the interface as real ColorRamp widgets.

    It will be created and determined by the 'material_name', so when a user creates a material with exactly this
    name first, it will mess things up.
    """

    material_name = "ColorBox_ColorRamps"

    @classmethod
    def get_or_create_colorbox_material(cls):
        try:
            return bpy.data.materials[cls.material_name]
        except KeyError:
            # create a new material, use nodes, but clear the nodes (default will create BSDF and output)
            # this is really a place to hold ColorRamp nodes, and nothing else
            material = bpy.data.materials.new(name=cls.material_name)
            material.use_nodes = True
            material.node_tree.nodes.clear()
            material.use_fake_user = True
            return material

    @classmethod
    def get_new_node_location(cls):
        material = cls.get_or_create_colorbox_material()

        locations = [node.location for node in material.node_tree.nodes]
        if not locations:
            return -600, 300

        max_node_location = max(locations, key=sort_key)
        new_x = max_node_location.x + 300
        new_y = max_node_location.y
        if new_x == -600 + 300 * 6:
            new_x = -600
            new_y = new_y - 300

        return new_x, new_y

    @classmethod
    def add_node(cls, ramp_data):
        material = cls.get_or_create_colorbox_material()
        # material_name = material.name
        # material.use_nodes = True

        color_ramp_node = material.node_tree.nodes.new("ShaderNodeValToRGB")
        color_ramp_node.location = cls.get_new_node_location()

        # TODO: fix this!
        if isinstance(ramp_data, str):
            ramp_data = json.loads(ramp_data)

        build_color_ramp_node(
            color_ramp_node=color_ramp_node, ramp_data=ramp_data
        )

        return color_ramp_node.name

    @classmethod
    def remove_node(cls, node_name):
        material = cls.get_or_create_colorbox_material()
        node_to_remove = material.node_tree.nodes[node_name]
        material.node_tree.nodes.remove(node_to_remove)

    @classmethod
    def remove_all_nodes(cls):
        material = cls.get_or_create_colorbox_material()
        for node in material.node_tree.nodes:
            material.node_tree.nodes.remove(node)

    @classmethod
    def get_node_by_name(cls, node_name):
        try:
            material = cls.get_or_create_colorbox_material()
            return material.node_tree.nodes[node_name]
        except KeyError as e:
            print(e)
    @classmethod
    def get_node_dict(cls, node_name):
        node = cls.get_node_by_name(node_name)
        return get_color_ramp_node_dict(color_ramp_node=node)
