import math


def get_hex_code(color):
    col = [int(c * 255) for c in color[:3]]  # math.ceil could work too
    # def clamp(x):
    #     return max(0, min(x, 255))
    # return "#{0:02x}{1:02x}{2:02x}".format(clamp(col[0]), clamp(col[1]), clamp(col[2]))
    return "#%02x%02x%02x" % (col[0], col[1], col[2])


def round_color_tuple(color_tuple, precision=4):
    return tuple((round(cv, precision) for cv in color_tuple))


def correct_gamma(c, gamma_value=2.4):
    if c < 0.0031308:
        srgb = 0.0 if c < 0.0 else c * 12.92
    else:
        srgb = 1.055 * math.pow(c, 1.0 / gamma_value) - 0.055

    return srgb


def inverse_gamma(r, gamma_value=2.4):
    if r < 0.04045:
        return r / 12.92 if r > 0.0 else 0.0

    return ((r + 0.055) / 1.055) ** gamma_value


class PixelIterator:
    """
    Iterate over blender Image pixel data.
    For small images, loop over every single pixel.
    For images larger than 1000 pixels, only consider 33 x 33 = 999 pixels.
    These 999 pixels are chosen from an 'equally distributed grid'.

    Return the color values as a Color tuple of floats.
    The color tuple always contains 4 rounded values (including the gamma value that will be 1 by default).
    """

    def __init__(self, image, analyse_all_pixels=False):
        self.analyse_all_pixels = analyse_all_pixels
        self.channel_count = image.channels
        self.width, self.height = image.size
        self.pixels = image.pixels[
            :
        ]  # way faster to use a Python list, instead of the Blender pixel datatype

        assert self.channel_count in [
            3,
            4,
        ], "PixelIterator expects a channel count of 3 or 4"
        assert self.width * self.height * self.channel_count == len(
            self.pixels
        ), f"PixelIterator: width x height * channel_count != len(pixels) for Image {image.path}"

    def get_color(self, start_index):
        r, g, b, *a = self.pixels[
            start_index : start_index + self.channel_count
        ]

        # a is a list of length 0 (when there are 3 channels) or length 1 (when there are 4 channels).
        a = (
            a[0] if len(a) == 1 else 1
        )  # use 1 as the default value for every pixel when there are 3 channels
        return round_color_tuple((r, g, b, a))

    def __iter__(self):
        if self.analyse_all_pixels:
            for x in range(self.width):
                for y in range(self.height):
                    start_index = (y * self.width + x) * self.channel_count
                    yield self.get_color(start_index=start_index)
        else:
            step = 45  # 45 * 45 = 2025 so this gives a ~2000 pixel sample
            x_step = self.width / step
            y_step = self.height / step
            half_x_step = x_step / 2
            half_y_step = y_step / 2

            for x in range(step):
                for y in range(step):
                    x_co = int(x * x_step + half_x_step)
                    y_co = int(y * y_step + half_y_step)
                    start_index = x_co * y_co * self.channel_count
                    yield self.get_color(start_index=start_index)

            # import random
            # for i in range(1000):
            #     x_co = random.randint(2, self.width - 2)
            #     y_co = random.randint(2, self.height - 2)
            #
            #     start_index = x_co * y_co * self.channel_count
            #     yield self.get_color(start_index=start_index)
