import json
import bpy
from pathlib import Path

from .color_ramp import ColorBoxMaterial
from .image import round_color_tuple


def set_color_box_from_list(color_box_list, context=None):
    scene = context.scene if context else bpy.context.scene

    if color_box_list:
        scene.color_box.clear()

    show_collection = True
    for collection_name, palettes in color_box_list.get(
        "palette_collections", []
    ):
        new_collection = scene.color_box.add()
        new_collection.name = collection_name
        new_collection.show = show_collection
        show_collection = False  # so only the first one added will be opened, all the other wil be closed.

        for palette_name, weighted_colors in palettes:
            new_palette = new_collection.palettes.add()
            new_palette.name = palette_name

            for color, weight in weighted_colors:
                if len(color) == 3:
                    # add 1 for the alpha value if is was a RGB color
                    color.append(1)

                new_color = new_palette.colors.add()
                new_color.color = tuple(color)
                new_color.weight = weight

    color_ramps_data = color_box_list.get("color_ramp_collections", [])
    if color_ramps_data:
        scene.color_box_ramps.clear()
        ColorBoxMaterial.remove_all_nodes()

    for collection_name, color_ramps in color_ramps_data:
        new_collection = scene.color_box_ramps.add()
        new_collection.name = collection_name

        for index, ramp_data_json in enumerate(color_ramps):
            color_ramp_node_name = ColorBoxMaterial.add_node(
                ramp_data=ramp_data_json
            )

            new_ramp = new_collection.ramps.add()
            new_ramp.name = color_ramp_node_name
            # new_ramp.json_data = ramp_data_json


def get_color_box_data_path() -> Path:
    return Path(bpy.utils.user_resource("CONFIG")) / Path("color_box.json")


def export_color_box(data_path, context):
    scene = context.scene if context else bpy.context.scene

    palette_collections = []
    for collection in scene.color_box:
        palettes = [
            [
                palette.name,
                [
                    (round_color_tuple(c.color), c.weight)
                    for c in palette.colors
                ],
            ]
            for palette in collection.palettes
        ]
        palette_collections.append([collection.name, palettes])

    ramp_collections = []
    for collection in scene.color_box_ramps:
        color_ramps = [
            ColorBoxMaterial.get_node_dict(node_name=ramp.name)
            for ramp in collection.ramps
        ]
        ramp_collections.append([collection.name, color_ramps])

    color_box = dict(
        palette_collections=palette_collections,
        color_ramp_collections=ramp_collections,
    )

    with open(data_path, "w") as outfile:
        json.dump(color_box, outfile)  # , indent=2)


def import_color_box(data_path, context=None):
    with open(data_path) as color_list_data:
        set_color_box_from_list(json.load(color_list_data), context=context)
