"""
This was copied from an operator. Some parts are left unchanged, so it could be reverted and used as an operator again
"""
from math import ceil

from .color_ramp import ColorBoxMaterial
from .image import inverse_gamma, correct_gamma


def update_pagination(scene, palettes):
    per_page = scene.color_box_palette_per_page
    palette_amount = len(palettes)
    scene.color_box_total_palettes = palette_amount

    if palette_amount > 0:
        scene.color_box_palette_page_amount = ceil(palette_amount / per_page)
    else:
        scene.color_box_palette_page_amount = 0


def add_palette(
    scene,
    palette_name,
    colors,
    collection_index,
    weights=None,
    open_collection=False,
):
    """
    A palette is one row in the Colorbox.
    It has a name, and a few colors, where the colors could have a weight.
    """
    curr_collection = scene.color_box[collection_index]
    palettes = curr_collection.palettes

    new_palette = palettes.add()
    new_palette.name = palette_name

    # when a new color was added, the collection can specifically set to be open
    if open_collection:
        curr_collection.show = True
    for index, _color in enumerate(colors):
        new_color = new_palette.colors.add()
        new_color.color = _color
        if weights and len(weights) > index:
            new_color.weight = weights[index]

    update_pagination(scene, palettes)

    return new_palette


def palette_to_ramp(
    palette,
    color_ramps,
    collection_to_index=None,
    row_to_index=None,
    new_collection_name=None,
):
    number_of_colors = len(palette.colors)
    if number_of_colors == 0:
        return
    elif number_of_colors == 1:
        step = 1
    else:
        step = 1 / (
            number_of_colors - 1
        )  # when there are 3 colors: add at 0.0, 0.5 and 1.0

    ramp_data = {
        "name": palette.name,
        "elements": [
            (
                index,
                index * step,
                tuple(
                    inverse_gamma(r, gamma_value=2.4) for r in color.color[:3]
                )
                + (color.color[3],),
                color.color[3],  # alpha
            )
            for index, color in enumerate(palette.colors)
        ],
    }
    new_name = ColorBoxMaterial.add_node(ramp_data=ramp_data)

    # add this new name in the interface
    if collection_to_index is None:
        new_collection = color_ramps.add()
        new_collection.name = (
            new_collection_name if new_collection_name else "New Collection"
        )
        collection_ramps = color_ramps[-1].ramps
    else:
        collection_ramps = color_ramps[collection_to_index].ramps

    new_ramp = collection_ramps.add()
    new_ramp.name = new_name

    if row_to_index is not None:
        row_from = (
            len(collection_ramps) - 1
        )  # the row was added as the last index
        color_ramps[collection_to_index].ramps.move(row_from, row_to_index)


def determine_collection_row(scene, self, collection_type_to=0):
    col_name = "color_box" if collection_type_to == 0 else "color_box_ramps"
    item_name = "palettes" if collection_type_to == 0 else "ramps"

    for col_index, collection in enumerate(getattr(scene, col_name)):
        for row_index, row in enumerate(getattr(collection, item_name)):
            if row.selected and row == self:
                return collection_type_to, col_index, row_index

        if collection == self:
            return collection_type_to, col_index, -1

    return -1, -1, -1


def ramp_to_palette(
    context, collection_from, row_from, collection_to, row_to=None
):
    scene = context.scene
    color_ramp = scene.color_box_ramps[collection_from].ramps[row_from]
    node_dict = ColorBoxMaterial.get_node_dict(node_name=color_ramp.name)

    colors = [element[2] for element in node_dict["elements"]]

    add_palette(
        scene,
        palette_name=color_ramp.name,
        collection_index=collection_to,
        colors=[
            tuple(correct_gamma(r, gamma_value=2.4) for r in color[:3])
            + (color[3],)
            for color in colors
        ],
    )

    if row_to is not None and row_to != -1:
        row_from = len(scene.color_box[collection_to].palettes) - 1
        scene.color_box[collection_to].palettes.move(row_from, row_to)


class MoveRow:
    """Move this row by clicking on this icon and then click on the icon where it should move to."""

    def __init__(self, thing, context, collection_type_to):
        scene = context.scene

        if not thing.selected:
            scene.color_box_selected_row = (-1, -1, -1)
            return

        move_from = tuple(scene.color_box_selected_row)
        move_to = determine_collection_row(
            scene, thing, collection_type_to=collection_type_to
        )
        if move_from == (-1, -1, -1):
            scene.color_box_selected_row = move_to
            return

        # the row was selected, and there is another row present, so it can be moved
        self.collection_type_to = (
            collection_type_to  # 0 for type ColorPalette, 1 for type ColorRamp
        )
        self.collection_to = move_to[1]
        self.row_to = move_to[2]

        try:
            self.execute(context, from_indices=move_from)
        except AttributeError as e:
            # this was originally made as an Operator, and used the self.report to log errors
            pass

        # deselect everything
        for collection in scene.color_box:
            for row in collection.palettes:
                row.selected = False
            collection.selected = False

        for collection in scene.color_box_ramps:
            for ramp in collection.ramps:
                ramp.selected = False
            collection.selected = False

        scene.color_box_selected_row = (-1, -1, -1)

    def add_copy_to_collection(self, collection, collection_from, row_from):
        """
        When moving between different collections, an item has to be re created in the new collection.
        After creation, it will be removed from the previous collection.

        The index of the new item will be returned, in case it also has to move to another place in the collection.
        """
        if self.collection_type_to == 0:
            from_row = collection[collection_from].palettes[row_from]
            new_item = collection[self.collection_to].palettes.add()
            new_item_index = len(collection[self.collection_to].palettes) - 1
            new_item.name = from_row.name

            for color in from_row.colors:
                new_color = new_item.colors.add()
                new_color.weight = color.weight
                new_color.color = color.color

            # remove the original row
            collection[collection_from].palettes.remove(row_from)

        else:
            from_row = collection[collection_from].ramps[row_from]
            new_item = collection[self.collection_to].ramps.add()
            new_item_index = len(collection[self.collection_to].ramps) - 1
            new_item.name = from_row.name

            # remove the original row
            collection[collection_from].ramps.remove(row_from)

        return new_item_index

    def convert_to_different_type(
        self, context, collection_type_from, collection_from, row_from
    ):
        """
        A color ramp can be converted into a color row, and vice versa.
        In this case, do not delete the source item.
        So it is not a move, it is a creation of a new item.
        """
        if not (collection_from >= 0 and row_from >= 0):
            self.report({"ERROR"}, f"This move is not possible.")
            return

        if collection_type_from == 0 and self.collection_type_to == 1:
            palette = context.scene.color_box[collection_from].palettes[
                row_from
            ]
            palette_to_ramp(
                palette,
                color_ramps=context.scene.color_box_ramps,
                collection_to_index=self.collection_to,
                row_to_index=self.row_to,
            )

        elif collection_type_from == 1 and self.collection_type_to == 0:
            ramp_to_palette(
                context,
                collection_from,
                row_from,
                collection_to=self.collection_to,
                row_to=self.row_to,
            )

        context.area.tag_redraw()

    def execute(self, context, from_indices):
        """
        Move a color row or a collection.
        """
        to_indices = (self.collection_type_to, self.collection_to, self.row_to)
        collection_type_from, collection_from, row_from = from_indices

        collection = (
            context.scene.color_box
            if collection_type_from == 0
            else context.scene.color_box_ramps
        )

        # a vector was saved already, use the indices to determined what has to be moved to where
        if collection_type_from != self.collection_type_to:
            self.convert_to_different_type(
                context, collection_type_from, collection_from, row_from
            )
        elif self.collection_type_to not in [0, 1]:
            self.report(
                {"ERROR"}, f"ColorBox move with invalid collection type."
            )
        elif from_indices == to_indices:
            # do nothing, but reset the context.scene.color_box_selected_row before returning FINISHED
            pass
        elif row_from == self.row_to == -1:
            # move a collection (can be a color_box collection, or color_ramp collection
            if collection_from != self.collection_to:
                collection.move(collection_from, self.collection_to)
                context.area.tag_redraw()
        elif (
            self.row_to == -1 and row_from >= 0
        ):  # collection to a row is not allowed
            # move a specific row to a collection means: adding it to the collection (will be the last item)
            self.add_copy_to_collection(collection, collection_from, row_from)
            context.area.tag_redraw()
        elif self.row_to == -1 or row_from == -1:
            self.report({"ERROR"}, f"Collection to a row is not allowed.")
        else:
            # move a single row
            if collection_from != self.collection_to:
                new_item_index = self.add_copy_to_collection(
                    collection, collection_from, row_from
                )
                # update the indices so the row will be moved correctly
                row_from = new_item_index

            # within a collection, move one item to a new index.
            item_name = "palettes" if collection_type_from == 0 else "ramps"
            getattr(collection[self.collection_to], item_name).move(
                row_from, self.row_to
            )
            context.area.tag_redraw()
