from math import pi, cos, sin
import gpu
from gpu_extras.batch import batch_for_shader
import blf


class CornerTypes:
    """
    Indexes work counter clockwise, starting at top right. This is useful for drawing rounded corners:
    to determine which quarter of a circle should be drawn.

    tl 1 0 tr
    bl 2 3 br
    """

    tr = 0
    tl = 1
    bl = 2
    br = 3


def draw_corner(x, y, color, segments, radius, corner_type):
    coords = []
    one_fourth = int(segments / 4)
    m = (1.0 / (segments - 1)) * (pi * 2)

    for p in range(
        corner_type * one_fourth - 1, (corner_type + 1) * one_fourth + 1
    ):
        p1 = x + cos(m * p) * radius
        p2 = y + sin(m * p) * radius
        coords.append((p1, p2))

    shader = gpu.shader.from_builtin("UNIFORM_COLOR")
    batch = batch_for_shader(shader, "TRI_FAN", {"pos": coords})

    shader.bind()
    shader.uniform_float("color", color)
    batch.draw(shader)

    # Draw one more triangle in the corner
    vertices = (
        (x, y),
        (x, y + radius if corner_type < 2 else y - radius),
        (x + radius if corner_type in [0, 3] else x - radius, y),
    )
    draw_filled_triangle(vertices=vertices, color=color)


def draw_filled_rectangle(start_x, start_y, width, height, color):
    vertices = (
        (start_x, start_y),
        (start_x + width, start_y),
        (start_x, start_y + height),
        (start_x + width, start_y + height),
    )

    indices = ((0, 1, 2), (2, 1, 3))

    shader = gpu.shader.from_builtin("UNIFORM_COLOR")
    batch = batch_for_shader(
        shader, "TRIS", {"pos": vertices}, indices=indices
    )

    shader.bind()
    shader.uniform_float("color", color)
    batch.draw(shader)


def draw_filled_triangle(vertices, color):
    shader = gpu.shader.from_builtin("UNIFORM_COLOR")
    batch = batch_for_shader(shader, "TRIS", {"pos": vertices})
    shader.bind()
    shader.uniform_float("color", color)
    batch.draw(shader)


def draw_filled_rounded_rectangle(
    start_x,
    start_y,
    width,
    height,
    color,
    rounding_radius=10,
    segments=24,
):
    """
    Draw 2 small bars (top and bottom) and a large middle section, something like this
       ____
    __|    |__
    |         |
    |__     __|
      |___|

    Then draw circle segments in all 4 corners
    """
    #
    # top and bottom bar with a height that equals the rounding radius
    #
    x_co = start_x + rounding_radius
    bar_width = width - 2 * rounding_radius
    draw_filled_rectangle(x_co, start_y, bar_width, rounding_radius, color)
    draw_filled_rectangle(
        x_co,
        start_y + height - rounding_radius,
        bar_width,
        rounding_radius,
        color,
    )

    #
    # middle section
    #
    subtract_height = 2
    draw_filled_rectangle(
        start_x,
        start_y + rounding_radius,
        width,
        height - subtract_height * rounding_radius,
        color,
    )

    #
    # Draw the 4 rounded corners
    #
    for (x_co, y_co, corner_type) in [
        (
            start_x + width - rounding_radius,
            start_y + rounding_radius,
            CornerTypes.br,
        ),
        (start_x + rounding_radius, start_y + rounding_radius, CornerTypes.bl),
        (
            start_x + rounding_radius,
            start_y + height - rounding_radius,
            CornerTypes.tl,
        ),
        (
            start_x + width - rounding_radius,
            start_y + height - rounding_radius,
            CornerTypes.tr,
        ),
    ]:
        draw_corner(
            x=x_co,
            y=y_co,
            color=color,
            segments=segments,
            radius=rounding_radius,
            corner_type=corner_type,
        )


def draw_rounded_border(
    start_x, start_y, width, height, color, rounding_radius=10, segments=24
):
    """
    Draw four rectangles, that form a square minus the corner points.
    Then draw circle segments at all 4 corners
    """
    # Top and bottom bar
    draw_filled_rectangle(
        start_x + rounding_radius,
        start_y,
        width - 2 * rounding_radius,
        rounding_radius,
        color,
    )
    draw_filled_rectangle(
        start_x + rounding_radius,
        start_y + height - rounding_radius,
        width - 2 * rounding_radius,
        rounding_radius,
        color,
    )
    # Left bar
    draw_filled_rectangle(
        start_x,
        start_y + rounding_radius,
        rounding_radius,
        height - 2 * rounding_radius,
        color,
    )
    # Right bar
    draw_filled_rectangle(
        start_x + width - rounding_radius,
        start_y + rounding_radius,
        rounding_radius,
        height - 2 * rounding_radius,
        color,
    )

    for x_diff, y_diff, corner_type in [
        (width - rounding_radius, rounding_radius, CornerTypes.br),
        (rounding_radius, rounding_radius, CornerTypes.bl),
        (rounding_radius, height - rounding_radius, CornerTypes.tl),
        (width - rounding_radius, height - rounding_radius, CornerTypes.tr),
    ]:
        draw_corner(
            x=start_x + x_diff,
            y=start_y + y_diff,
            color=color,
            segments=segments,
            radius=rounding_radius,
            corner_type=corner_type,
        )


def draw_text(
    x_co, y_co, font_size, text, text_color=None, center_total_width=None
):
    if text_color is None:
        text_color = (1.0, 1.0, 1.0)

    blf.size(0, font_size)
    if center_total_width is not None:
        # calculate an offset in pixels for the width, so the text will be centered
        text_width = blf.dimensions(0, text)[0]
        # take half of the free pixels (space on left and right), which is the offset to add to the x coordinate
        x_co += int((center_total_width - text_width) / 2)

    # font_id = 0  # ??
    blf.color(
        0, text_color[0], text_color[1], text_color[2], 1.0
    )  # fontid, rgba
    blf.position(0, x_co, y_co, 0)
    blf.draw(0, text)


def draw_2d_line(coords, color=None):
    if color is None:
        color = (0, 0, 0, 1)

    new_coords = []
    len_coords = len(coords)
    for index in range(len_coords):
        new_coords.append(coords[index])
        new_coords.append(coords[(index + 1) % len_coords])

    shader = gpu.shader.from_builtin("UNIFORM_COLOR")
    batch = batch_for_shader(
        shader=shader, type="LINES", content={"pos": new_coords}
    )

    shader.bind()
    shader.uniform_float("color", color)
    batch.draw(shader)


def draw_image(image, location, width, height):
    shader = gpu.shader.from_builtin("IMAGE")
    batch = batch_for_shader(
        shader,
        "TRI_FAN",
        {
            "pos": (
                location,
                (location[0] + width, location[1]),
                (location[0] + width, location[1] + height),
                (location[0], location[1] + height),
            ),
            "texCoord": ((0, 0), (1, 0), (1, 1), (0, 1)),
        },
    )

    try:
        # 使用 gpu 模块替代 bgl 模块
        texture = gpu.texture.from_image(image)
        
        shader.bind()
        shader.uniform_sampler("image", texture)
        
        gpu.state.blend_set('ALPHA')  # 启用透明混合
        batch.draw(shader)
        gpu.state.blend_set('NONE')   # 关闭混合
    except Exception as e:
        print("Could not display the logo with the gpu module")
        print(repr(e))


def draw_rounded_border_perforated(
    start_x, start_y, width, height, color, rounding_radius=10, segments=24
):
    """
    Draw four rectangles, that form a square minus the corner points.
    Then draw circle segments at all 4 corners
    """
    # Top and bottom bar
    draw_filled_rectangle(
        start_x + rounding_radius,
        start_y,
        width - 2 * rounding_radius,
        rounding_radius,
        color,
    )
    draw_filled_rectangle(
        start_x + rounding_radius,
        start_y + height - rounding_radius,
        width - 2 * rounding_radius,
        rounding_radius,
        color,
    )

    bar_height = height - 2 * rounding_radius
    new_length = bar_height / 3.5

    bar_y = start_y + rounding_radius
    bar_y_2 = bar_y + bar_height - new_length

    for _start_x, _start_y in [
        (start_x, bar_y),
        (start_x, bar_y_2),
        (start_x + width - rounding_radius, bar_y),
        (start_x + width - rounding_radius, bar_y_2),
    ]:
        draw_filled_rectangle(
            _start_x, _start_y, rounding_radius, new_length, color
        )

    for x_diff, y_diff, corner_type in [
        (width - rounding_radius, rounding_radius, CornerTypes.br),
        (rounding_radius, rounding_radius, CornerTypes.bl),
        (rounding_radius, height - rounding_radius, CornerTypes.tl),
        (width - rounding_radius, height - rounding_radius, CornerTypes.tr),
    ]:
        draw_corner(
            x=start_x + x_diff,
            y=start_y + y_diff,
            color=color,
            segments=segments,
            radius=rounding_radius,
            corner_type=corner_type,
        )
